//
// LILC.C   (c) Copyright 2000, 2001 by Jim Steuert
//
//
//  Revision 1.03: Added moveto,clrscr,font,str,
//                 and line romcall function names.
//
//                 Improved examples with an input
//                 subroutine which allow backspaces.
//
//                 FIXED BUG IN txt289.c which caused
//                 crashes eventually.
//
//                 Confirmed that lilc works on TI92P.
//                 I really love my ti92. The keyboard
//                 makes it great for programming on
//                 the calc itself.
//
//  Revision 1.02: First public release of source code
//                 Fixes divide /
//                 Implements rem (same as c %).
//                 Implements initializers.
//                 Implements labels and goto.
//
//  A small c-like compiler which
//  compiles on the ti89. See the readme.txt
//  included with this.
//
//  You can use this compiler for any purpose
//  whatsoever as long as my name is posted
//  on the final source code and/or is displayed
//  as part of the about or informational
//  screen of the resulting software, and also
//  provided that no liability is implied of
//  this source code or it's implementation
//  in other packages.
//
//  Warning: This code has NOT been thoroughly
//           tested, and should be evaluated
//           and fully tested for its final use.
//
//  Warning: This implements a small subset of
//           the c language. It does not presume
//           to implement all c features.
//
//
// To compile for TI89:
//
//   where include is the tigcc include lib with the
//      tigcclib includes
//      and inits.o is in the current directory
//
//   tigcc -O2 -Iinclude -DTI89_VERSION lilc.c inits.o
//
//
// To run on the ti89:
//
//     doorsos.89z must have been installed by typing doorsos()
//
//     then to run lilc type lilc() and hit enter
//     or else type  lilc("textfilename",4) and hit enter
//
//    where 3 means generate a listing
//          4 means compile-and-go
//
//
//
// To compile for the PC (X86) using the mingw gcc version
//
//   gcc -o lilc -DCMDLINE_VERSION  lilc.c
//
//
// To run on the PC (X86), type:
//
//    lilc  <textfile> [ 1-3 ]
//
//    which generates a listing

#ifdef TI89_VERSION

#include   "doors.h"
#include   "string.h"
#include   "stdio.h"
#include   "stdlib.h"
#include   "alloc.h"
#include   "kbd.h"
#include   "graph.h"
#include   "args.h"
int   _ti89;

#endif

#ifdef CMDLINE_VERSION

#include <stdio.h>
#include <string.h>

#endif


char *gettkn();
short int  parse_dataref();
short int  getaregno( char * regname, short * );
short int  getdregno( char * regname, short * );
short int  getromcall( char *functionname );

void emit2();
void emit4();
void emit6();

int compile_only;
int compile_lines;
int compile_listing;
int compile_and_go;


int  numgbls;
char gblname[100][32]; // name of gbl var
int  gbldim[100];      // array dimension
int  gbloffset[100];   // offset rel a5
int  gbltype[100];     // char=1/short=2/long=3
int  gblsize[100];     // 1, 2, 4
int  gblsign[100];     // signed or unsigned


int  nextaddrreg;        // cur addr reg a1-a5
char addrregname[8][32]; // name of n-th register
int  addrregtosize[8];   // size of pointee

int  nextdatareg;        // cur data reg d2-d7
char dataregname[8][32]; // name of n-th register
int  dataregsize[8];     // 1,2, or 4
int  datasign[8];        // 0 = unsigned, 1 = signed


char varname[100];
char basename[100];
int  varlen;
int  sign;
int  type;
int  length;
int  pointer;
int  gotname;
int  code;

#define EQ  10
#define NE  11
#define GE  12
#define LE  13
#define GT  14
#define LT  15


#define CHAR 1
#define SHORT 2
#define LONG  4
#define UNSIGNED 50
#define SIGNED   60
#define NONE     70

#define DATA_VAR  20
#define REG_VAR   21

short int  result_type;
short int  aregno;
short int  dregno;
long constant;
char aregname[100];
int  offset;
int  datatype;
int  datalen;
int  indexlen;

char *cursor;
char *call_cursor;
char token[1024];

int first_line;
char *stmt;
char *newcursor;


// INTERNAL INPUT FILE

FILE *stream;

// WHERE COMPILE EMITS LITERALS/CODE

char *a5_area;
unsigned char *codeptr;


// IMPORTANT INTERNAL CURSORS

int   a5_litoff;
int   a5_gbloff;

// INTERNAL CODE EMIT CURSOR

unsigned int   codeoff;


int rslt;


int  else_branch_ref;


#define IF   4001
#define FOR  4002
#define NO_STATEMENT 4003

int if_fix_offset;

int brace_level;
int last_statement_type;
int brace_type[20];
int brace_if_fix[20];

int continue_offset[20];

int for_level;
int numbreaks[20];
int break_fix[20][100];

short continue_off;
char  tempstr[24];


// USED BY emit2(), emit4(), emit6()

unsigned int opword;
unsigned short dispword;
unsigned short thirdword;


short int errcode;

short  numgotos;
short  goto_fix_offset[20];
char   goto_label[20][32];

short  numlabels;
short  labeloff[20];
char   label[20][32];


int in_function;
int in_function_arglist;
int function_a6_offset;

int   num_functions;
short function_code_offset[20];
char  function_name[20][32];
int   branch_around_function;
int   movem_a5_offset;

int   num_function_args;
int   function_movem_offset;

int   branch_around_offset;
unsigned short   movem_mask;


#ifdef TI89_VERSION

int _main(void)
{

#endif

#ifdef CMDLINE_VERSION

int main(argc,argv)
int argc;
char *argv[];
{

#endif

    unsigned short inchar;

    short shortarg;
    long  longarg;
    char *fp;
    char *pp;
    char *gp;

    char *fopenp;
    char *fgetcp;
    char *fgetsp;
    char *fclosep;
    char *fprintfp;
    char *fputcp;
    char *atoip;
    char *atolp;
    char *strtolp;
    char *strtoulp;

    unsigned long *dp;
    void (*mp)();
    char *xp;
    int compile_error;
    unsigned short *codep;
    int i;
    char filename[256];

//
// Note: These hex codes for the c-runtime routines are taken from
//       Zelkjo Juric's excellent c runtime
//       include files, which are part of the tigcclib.zip
//       package. Refer to that package for use.
//

    short fputcharcode[80] = {0x4E56,-20,0x48E7,0x1F3C,0x4240,
      0x1C2E,9,0x2A78,0xC8,0x206D,0x638,0x4E90,0x3800,0xE34C,0x5C44,0x266D,0xBC,
      0x486E,-18,0x206D,0x680,0x4E90,0x3A2E,-8,0x362E,-6,0x3F06,0x206D,0x640,
      0x4E90,0x3E00,0xC06,10,0x670E,0x3005,0xD047,0x4241,0x122B,2,0xB240,0x6C04,
      0x4245,0xD644,0x3003,0xD044,0x4241,0x122B,3,0xB240,0x6C10,0x4257,0x3F04,
      0x2F0B,0x2F0B,0x206D,0x62C,0x4E90,0x9644,0x3EBC,4,0x3F06,0x3F03,0x3F05,
      0xC06,10,0x6708,0x206D,0x690,0x4E90,0xDF57,0x206D,0x674,0x4E90,0x3006,
      0x4CEE,0x3CF8,-56,0x4E5E,0x4E75};




    short getscode[28] = {
      0x4e56,0x0000, // link a6,#0
      0x266e,0x0008, // movea.l 8(a6),a3
// loop:
      0x2078,0x00c8, // movea.l $c8,a0
      0x2068,0x0144, // movea.l 0x144(a0),a0
      0x4e90,        // jsr     (a0)   ngetchx
      0x0c40,0x000d, // cmpi.w #13,d0
      0x6600,0x0004, // bne +4
      0x700a,        // moveq #10,d0  got 0d => echo 0a
      0x3600,        // move.w d0,d3
      0x16c3,        // move.b d3,(a3)+\n");
      0x3f03,        // move.w d3,-(a7)
      0x206d,0x0004, // movea.l 4(a5),a0
      0x4e90,        // jsr (a0)      putchar
      0x548f,        // addq.l #2,a7
      0x0c43,0x000a, // cmpi.w #10,d3
      0x6600,0xffd8, // bne loop (-40)\n");
// done:
      0x4223,        // clr.b -(a3)
      0x4e5e,        // unlk a6
      0x4e75 };       // rts


    short printfcode[19] 
      = {0x2078,
         0x00C8,
         0x2068,
         0x014C,
         0x41E8,
         0x0020,
         0x3250,
         0x486F,
         0x000C,
         0x2F2F,
         0x000C,
         0x42A7,
         0x2F2F,
         0x0010,
         0x4EB1,
         0x8800,
         0x4FEF,
         0x0010,
         0x4E75};



short fopencode[195] = {0x4E56,0xFFCC,0x48E7,0x1E3C,
  0x246E,8,0x2C2E,12,0x49EE,0xFFCE,0x2246,0x1811,0x4245,0xC29,0x62,1,0x6708,
  0xC29,0x62,2,0x6602,0x7A01,0x2078,0xC8,0x4878,18,0x2068,0x288,0x4E90,
  0x2A48,0x588F,0xBAFC,0,0x6700,0xAC,0x421C,0x6002,0x528C,0x101A,0x1880,
  0x66F8,0xC04,0x72,0x6706,0xC04,0x61,0x6658,0x7601,0x2478,0xC8,0x2F0C,
  0x206A,0x180,0x4E90,0x2F00,0x206A,0x1E4,0x4E90,0x2448,0x508F,0xB4FC,0,
  0x660A,0xC04,0x72,0x676C,0x7877,0x602E,0x2078,0xC8,0x302A,12,0x3B40,8,
  0x3F00,0x2068,0x264,0x4E90,0x2648,0x2078,0xC8,0x3F2D,8,0x2068,0x278,
  0x4E90,0x426D,14,0x3B40,16,0x588F,0xC04,0x77,0x6666,0x7602,0x367C,0xC8,
  0x2453,0x2F0C,0x206A,0x170,0x4E90,0x2F00,0x206A,0x1E4,0x4E90,0x2448,
  0x2053,0x4878,0x3E8,0x2068,0x240,0x4E90,0x3540,12,0x3B40,8,0x4FEF,12,
  0x6606,0x91C8,0x6000,0x88,0x2B7C,0,0x3E8,14,0x2078,0xC8,0x3F00,0x2068,
  0x264,0x4E90,0x2648,0x4A45,0x6704,0x4253,0x600E,0x26BC,5,1,0x277C,0x2000,
  0xE000,4,0x4280,0x3013,0xD08B,0x4A45,0x6702,0x5480,0xC04,0x61,0x6606,
  0x7602,0x2200,0x600C,0x220B,0x5A81,0x4A45,0x6704,0x220B,0x5481,0xB280,
  0x6604,0x43,32,0x2246,0xC29,0x2B,1,0x6708,0xC29,0x2B,2,0x6604,0x43,3,
  0x4A45,0x6704,0x43,0x40,0x3B43,10,0x2B4B,4,0x2A81,0x426D,12,
  0x204D,0x4CEE,0x3C78,0xFFAC,0x4E5E,0x4E75};



long fgetccode[36] = {0x4E560000,0x48E71820,0x246E0008,
  0x302A000A,0x3600EC4B,0x0A430001,0x2430001,0x8000004,0x660E0800,0x660C,
  0x400010,0x3540000A,0x70FF6052,0x382A000C,0x6C0A426A,0xC4240,0x10046042,
  0x8000005,0x66E62052,0x42441810,0x52920C44,0xD660A,0x4A436706,0x2F0A4EBA,
  0xFFA0206A,0x44280,0x3010D088,0x22124A43,0x66085480,0xB2806706,0x600AB280,
  0x6606006A,0x20000A,0x30044CEE,0x418FFF4,0x4E5E4E75};



long fgetscode[25] = {0x4E560000,
  0x48E71838,0x286E0008,0x282E000C,0x266E0012,0x2444362E,0x106024,
  0x2F0B4E94,0x588F0C40,0xFFFF671E,0xC40000D,0x660A082B,0x6000B,0x6602700A,
  0x14C00C00,0xA6706,0x53434A43,0x6ED64212,0xC40FFFF,0x6604B88A,0x67042004,
  0x60027000,0x20404CEE,0x1C18FFEC,0x4E5E4E75};



short fclosecode[35] = {0x4E56,0,0x48E7,0x1030,0x266E,8,
  0xB6FC,0,0x6604,0x70FF,0x6026,0x82B,4,11,0x56C3,0x4883,0x347C,0xC8,0x2052,
  0x3F2B,8,0x2068,0x27C,0x4E90,0x2052,0x2F0B,0x2068,0x28C,0x4E90,0x3003,
  0x4CEE,0xC08,0xFFF4,0x4E5E,0x4E75};


long fprintfcode[10] = {0x207800C8,0x2068014C,0x41E80020,0x3250486F,
  0x102F2F,0x102F2F,0x102F2F,0x104EB1,0x88004FEF,0x104E75};


long fputccode[71] = {0x4E560000,0x48E71C30,0x382E0008,
  0x266E000A,0x322B000A,0x3601EC4B,0xA430001,0x2430001,0x246B0004,
  0x2A0A0801,0x46706,0x70FF6000,0xE20801,0x1660E,0x410010,0x3741000A,
  0x70FF6000,0xCE3012,0x640000A,0x2800000,0xFFFFB0AB,0xE635C,0x347C00C8,
  0x20523F2B,0x82068,0x27C4E90,0x2052202B,0xE0680,0x3E8,0x2740000E,
  0x2F003F2B,0x82068,0x2744E90,0x508F4A40,0x660A006B,0x10000A,0x70FF607E,
  0x207800C8,0x3F2B0008,0x20680264,0x4E902748,0x42448,0x200A9085,0xD193548F,
  0x322B000A,0x8010005,0x67025252,0xC44000A,0x66064A43,0x6702780D,
  0x20531084,0x52930C44,0xD660E,0x4A43670A,0x2F0B3F3C,0x204EBA,0xFF204280,
  0x3012D08A,0x22134A43,0x66085480,0xB2806706,0x601AB280,0x6616006B,
  0x20000A,0x4A43670C,0x20534210,0x2053117C,0xFFE00001,0x30044CEE,0xC38FFEC,
  0x4E5E4E75};


long atoicode[25] = {0x4E560000,0x2F03206E,0x84241,
  0x42436002,0x52881010,0xC000020,0x67F60C00,0x2D6706,0xC0000AD,0x66085288,
  0x363CFFFF,0x60180C00,0x2B6612,0x5288600E,0x3001E748,0xD041D240,
  0xD2420641,0xFFD04242,0x14183002,0x640FFD0,0xC400009,0x63E23001,
  0x4A436702,0x4440261F,0x4E5E4E75};


long atolcode[31] = {0x4E560000,0x2F042F03,0x226E0008,
  0x42447400,0x60025289,0x10110C00,0x2067F6,0xC00002D,0x67060C00,0xAD6608,
  0x5289383C,0xFFFF6008,0xC00002B,0x66025289,0x42411219,0x30010640,
  0xFFD00C40,0x96224,0x76002002,0xD080E78A,0x2042D1C0,0x360141F0,0x38D02408,
  0x42411219,0x30010640,0xFFD00C40,0x963DE,0x20024A44,0x67024480,0x261F281F,
  0x4E5E4E75};


long strtolcode[91] = {0x4E56FFF8,0x48E71F3C,
  0x3C2E0010,0x266E0008,0x387C0000,0x4243161B,0x0C430020,0x67F80C43,
  0x2D6706,0xC4300AD,0x6606387C,0x16006,0xC43002B,0x66044243,0x161B4A46,
  0x67060C46,0x10661E,0xC430030,0x66181013,0xC000078,0x67060C00,0x58660A,
  0x4243162B,0x1548B,0x7C104A46,0x660A7C0A,0xC430030,0x66027C08,0x283C7FFF,
  0xFFFF3E0C,0x6706283C,0x80000000,0x34462204,0x200A2078,0xC82068,0xAA84E90,
  0x24412204,0x200A2078,0xC82068,0xAAC4E90,0x3D41FFFA,0x78004245,0x42AEFFFC,
  0x6038BC43,0x63784A45,0x6D0CB5C4,0x6508660A,0xB66EFFFA,0x63047AFF,
  0x601C7A01,0x2004C0C6,0x22044841,0xC2C64841,0x4241D081,0x28003D43,
  0xFFFED8AE,0xFFFC4243,0x161B3003,0x640FFD0,0xC400009,0x62043600,
  0x60B84242,0x32030641,0xFFBF0C41,0x19630C,0x30030640,0xFF9F0C40,0x196202,
  0x74014A42,0x67143003,0x640FFA9,0xC410019,0x62D03003,0x640FFC9,0x60C84A45,
  0x6C12283C,0x7FFFFFFF,0x3E0C670E,0x283C8000,0x6006,0x3E0C6702,0x44844AAE,
  0xC6716,0x4A45670A,0x538B2A6E,0xC2A8B,0x60082A6E,0xC2AAE,0x82004,
  0x4CDF3CF8,0x4E5E4E75};


long strtoulcode[73] = {0x4E560000,
  0x48E71F3C,0x2A6E0008,0x286E000C,0x3A2E0010,0x264D4243,0x161B0C43,
  0x2067F8,0xC43002B,0x66044243,0x161B4A45,0x67060C45,0x10661E,0xC430030,
  0x66181013,0xC000078,0x67060C00,0x58660A,0x4243162B,0x1548B,0x7A104A45,
  0x660A7A0A,0xC430030,0x66027A08,0x344572FF,0x200A2078,0xC82068,0xAA84E90,
  0x244172FF,0x200A2078,0xC82068,0xAAC4E90,0x3E0191C8,0x42447C00,0x6032BA43,
  0x63724A44,0x6D0AB5C8,0x65066608,0xBE436404,0x78FF6018,0x78012008,
  0xC0C52208,0x4841C2C5,0x48414241,0xD0812040,0x3C03D1C6,0x4243161B,
  0x30030640,0xFFD00C40,0x96204,0x360060BE,0x42423203,0x641FFBF,0xC410019,
  0x630C3003,0x640FF9F,0xC400019,0x62027401,0x4A426714,0x30030640,
  0xFFA90C41,0x1962D0,0x30030640,0xFFC960C8,0x4A446C04,0x307CFFFF,
  0xB8FC0000,0x670C4A44,0x6706538B,0x288B6002,0x288D2008,0x4CDF3CF8,
  0x4E5E4E75};



  long num;

#ifdef TI89_VERSION

  ESI argptr;
  int argtype;



   compile_only = 0;
   compile_lines = 0;
   compile_listing = 0;
   compile_and_go = 0;
   

   clrscr();

   InitArgPtr(argptr);

   argtype = GetArgType(argptr);

   if ( argtype != END_TAG )
   {
      if ( argtype != STR_TAG )
      {
        DrawStr(0,20,"First arg must be filename",A_REPLACE);
        DrawStr(0,30,"Usage: lilc(\"file\",1-4)",A_REPLACE);
        ngetchx();
		return(1);
      }

      strcpy( filename, GetStrnArg(argptr) );
      printf("file: %s\n",filename);
    
      argtype = GetArgType(argptr);
    
    
      if ( argtype != POSINT_TAG && argtype != NEGINT_TAG )
      {
        DrawStr(0,20,"Second arg must be 1-4",A_REPLACE);
        DrawStr(0,30,"Usage: lilc(\"file\",1-4)",A_REPLACE);
        ngetchx();
		return(1);
      }

      num=GetIntArg(argptr);
    
      if ( num > 4 || num < 1 )
      {
        DrawStr(0,20,"Second xx arg must be 1-4",A_REPLACE);
        DrawStr(0,30,"Usage: lilc(\"file\",1-4)",A_REPLACE);
        ngetchx();
		return(1);
      }

      if ( num == 1 )
         compile_only = 1;
      else if ( num == 2 )
         compile_lines = 1;
       else if ( num == 3 )
         compile_listing = 1;
       else if ( num == 4 )
         compile_and_go = 1;
       else
         compile_only = 1;
   }
   else
   {
       printf("Enter filename: ");
    
       gets( filename );
    
       ngetchx();

       printf("Enter:\n");
       printf(" 1: compile only\n");
       printf(" 2: compile w/lines\n");
       printf(" 3: compile w/listing\n");
       printf(" 4: compile and go\n");
   
       inchar = ngetchx();
   
       if ( inchar == '1' )
         compile_only = 1;
       else if ( inchar == '2')
         compile_lines = 1;
       else if ( inchar == '3')
         compile_listing = 1;
       else if ( inchar == '4')
         compile_and_go = 1;
       else
         compile_only = 1;
   }
#endif

#ifdef CMDLINE_VERSION

#define ngetchx()  

   compile_only = 0;
   compile_lines = 0;
   compile_listing = 0;
   compile_and_go = 0;
  
 
   if ( argc < 2 )
   {
      printf("Usage: lilc <filename> [ 1-3 ] \n");
      return(1);
   }

   strcpy( filename, argv[1] );

   num = 3; // default

   if ( argc == 3 )
      num = atoi( argv[2] );

   if ( num == 1 )
     compile_only = 1;
   else if ( num == 2 )
     compile_lines = 1;
   else if ( num == 3 )
     compile_listing = 1;
   else if ( num == 4 )
     compile_and_go = 1;
   else
     compile_only = 1;

#endif

    stream = fopen( filename , "r" );

    if ( stream == NULL)
    {
       printf("can't open file %s\n",filename);
       ngetchx();
       return(1);
    }

    fp = (char *) malloc( 200 );    
    memcpy( fp, &fputcharcode[0], 160 );

    pp = (char *) malloc( 200 );    
    memcpy( pp, &printfcode[0], 38 );

    gp = (char *) malloc( 100 );    
    memcpy( gp, &getscode[0], 56 );

    fopenp = (char *) malloc( 400 );    
    memcpy( fopenp, &fopencode[0], 195*2 );

    fgetcp = (char *) malloc( 200 );    
    memcpy( fgetcp, &fgetccode[0], 36*4 );

    fgetsp = (char *) malloc( 200 );    
    memcpy( fgetsp, &fgetscode[0], 25*4 );

    fclosep = (char *) malloc( 100 );    
    memcpy( fclosep, &fclosecode[0], 35*2 );

    fprintfp = (char *) malloc(60);
    memcpy( fprintfp, &fprintfcode[0], 10*4 );

    fputcp = (char *) malloc(400);
    memcpy( fputcp, &fputccode[0], 71*4 );

    atoip = (char *) malloc(120);
    memcpy( atoip, &atoicode[0], 25*4 );

    atolp = (char *) malloc(150);
    memcpy( atolp, &atolcode[0], 31*4 );

    strtolp = (char *) malloc(380);
    memcpy( strtolp, &strtolcode[0], 91*4 );

    strtoulp = (char *) malloc(300);
    memcpy( strtoulp, &strtoulcode[0], 73*4 );

    a5_area = (char *) malloc( 4096 ); // data area

    codeptr = (unsigned char *) malloc( 4096 );
    codeoff = 0;
    a5_litoff = 512;
    a5_gbloff = 1024;

    compile_error = compile();

    if ( compile_error != 0 )
    {
       printf("compile error %d\n",compile_error);
       ngetchx();
       return(1);
    }
    else
    {
       printf("compile ok\n");
       ngetchx();
    }

#ifdef TI89_VERSION
    clrscr();
#endif

    dp = (unsigned long *) a5_area;

    dp[0] = (unsigned long) pp;
    dp[1] = (unsigned long) fp;
    dp[2] = (unsigned long) gp;
    dp[3] = (unsigned long) fopenp;
    dp[4] = (unsigned long) fgetcp;
    dp[5] = (unsigned long) fgetsp;
    dp[6] = (unsigned long) fclosep;
    dp[7] = (unsigned long) fprintfp;
    dp[8] = (unsigned long) fputcp;
    dp[9] = (unsigned long) atoip;
    dp[10] = (unsigned long) atolp;
    dp[11] = (unsigned long) strtolp;
    dp[12] = (unsigned long) strtoulp;

    mp = (void (*)()) codeptr;

#ifdef TI89_VERSION
    (mp)(dp);
#endif

    printf("Done\n");
    ngetchx();

#ifdef CMDLINE_VERSION
    codep = ( unsigned short *) codeptr;
    printf("%d shorts:\n", (codeoff/2) );

    for ( i=0; i<(codeoff/2); i++)
    {
       printf("0x%04x,",codep[i]);
       if ( (i & 0x7) == 7 )
       {
         printf("\n");
       }
    }
#endif

    ngetchx();

    free(codeptr);
    free(fp);
    free(pp);
    free(gp);

    free(fopenp);
    free(fgetcp);
    free(fgetsp);
    free(fclosep);
    free(fprintfp);
    free(fputcp);
    free(atoip);
    free(atolp);
    free(strtolp);
    free(strtoulp);

    free(a5_area);

    return(0);

}












int compile()
{

int  error;
int  len;
char line[1024];
char *cptr;
char *ap;
char *fp;
char *dimp;
int  i;
int  j;
int  dim;
int  lineno;
char *xp;
int  regno;
int  data_or_reg;
char *cp;
unsigned short *codep;

int  rslt_type;
int  rslt_aregno;
int  rslt_dregno;
long  rslt_constant;
int  rslt_offset;
int  rslt_datatype;
int  rslt_datalen;
int  rslt_indexlen;

int  left_type;
int  left_aregno;
int  left_dregno;
long  left_constant;
int  left_offset;
int  left_datatype;
int  left_datalen;
int  left_indexlen;

int  operator;

int  right_type;
int  right_aregno;
int  right_dregno;
long  right_constant;
int  right_offset;
int  right_datatype;
int  right_datalen;
int  right_indexlen;


int  comparison;

int    if_dregno;
short  if_dreglen;
int    compare_dregno;
short  compare_dreglen;
short  else_mask;




char   opname[32];
short  xreglen;
unsigned long immediate;

#define  NOOP 0
#define  XOR 1
#define  ADD 2
#define  SUB 3
#define  AND 4
#define  OR  5
#define  MUL 6
#define  DIV 7
#define  SHR 8
#define  SHL 9
#define  ROL 10
#define  ROR 11
#define  ASN 12
#define  REM 13



int type;

unsigned int areg;
unsigned int dreg;
unsigned long con;
unsigned short shortcon;
unsigned int dtype;

unsigned int dlen;
unsigned int ilen;
unsigned int off;

unsigned short len_mask;
unsigned short op_mask;
unsigned short dest_mask;
unsigned short src_regno;
unsigned short imm_mask;

int arg;

char condition[10];

int  fix_offset;



int  numargs;
int  arg_type[12];
int  arg_aregno[12];
int  arg_dregno[12];
long arg_constant[12];
int  arg_offset[12];
int  arg_datatype[12];
int  arg_datalen[12];
int  arg_indexlen[12];




short  for_dregno;
short  for_dreglen;
long   for_increment;

int limit;
int limit_dregno;
short limit_dreglen;

int breaks;
int breakno;

int total_len;
int romcall;
char functionname[256];

int set_aregno;
int set_dregno;
short set_dreglen;
short set_areglen;
int set_flag;

int opt_flag;
int dest_reg;

long *longp;

char  *curp;
short *curshortp;
long  *curlongp;

long  longval;

int   found_label;


//*****************************

in_function = 0;
in_function_arglist = 0;
function_a6_offset = 0;
num_function_args = 0;
function_movem_offset = 0;
movem_a5_offset = 64;
num_functions = 0;


//
// CODE PREAMBLE
// 0x4E56,0x0000, //link.w %a6,#0
// 0x2F0D,        //move.l %a5,-(%sp)
// 0x2a6E,0x0008, //move.l 8(%a6),%a5       data area pointer
//

codep = (unsigned short *) codeptr;
codep[0] = 0x4e56;
codep[1] = 0x0000;
codep[2] = 0x48e7;
codep[3] = 0x3ffc;
codep[4] = 0x2a6e;
codep[5] = 0x0008;
codeoff = 12;

longp = (long *) a5_area;

for (i=0;i<1024;i++)
{
   *longp++ = 0;
}


lineno = 0;


numgbls  = 0;
numargs  = 0;
nextaddrreg = 1;  // cur addr reg a1-a4
nextdatareg = 2;  // cur data reg d2-d7

else_branch_ref = 0;


brace_level = 0;
for_level = 0;
last_statement_type = NO_STATEMENT;

numlabels = 0;
numgotos  = 0;
first_line = 1;


for(;;)  /* FOR LOOP OVER LINES OF INPUT */
{
    //
    // get next statement
    //
    for(;;) /* skip comment lines */
    {
      if (first_line != 1 )
      {
         newcursor = strstr(cursor,";");
      }

      if (     first_line == 1 
           || (newcursor = strstr(cursor,";")) == NULL
         )
      {
        cptr = fgets( line , 500 , stream );
        lineno++;
        first_line = 0;

        if ( cptr == NULL || ((short) memcmp(line,"end",3)) == 0 )
        {
           //
           // END OF COMPILE
           //

           //
           // CODE POSTAMBLE
           // rom call 51  getchar
           //        {0x2078,0x00C8, //movea.l  abs.w,a0
           //         0x2068,0x0144, //movea.l  d(a0),a0 d = 4*0x51
           //         0x4E90,        //jsr (a0)
           //         0x2A6E,0xFFFC, //movea.l -4(%a6),a5
           //         0x4E5E,        //unlk %a6
           //         0x4E75};       //rts
           //

           cp = codeptr + codeoff;
           codep = (unsigned short *) cp;
           codep[0] = 0x2078;
           codep[1] = 0x00c8;
           codep[2] = 0x2068;
           codep[3] = 0x0144;
           codep[4] = 0x4e90;
           codep[5] = 0x4cee;
           codep[6] = 0x3ffc;
           codep[7] = 0xffd0;  // -48(a6)
           codep[8] = 0x4e5e;
           codep[9] = 0x4e75;
           codeoff += 20;

           // now fix up gotos

           for (i=0;i<numgotos;i++)  // fix each goto
           {
              found_label = 0;

              for (j=0;j<numlabels;j++) // find it's label
              {
                 if ( ((short) strcmp(goto_label[i],label[j])) == 0 ) //found label
                 {
                    found_label = 1;
       
                    cp = codeptr+goto_fix_offset[i];
                    codep = (unsigned short *) cp;
       
                    *codep = labeloff[j]-goto_fix_offset[i];
                    if ( compile_listing == 1 )   
                       printf("goto: code[ 0x%04x ] = 0x%04x\n",
                              goto_fix_offset[i],*codep);   
                    break;
                 }
             }
             if ( found_label == 0 )
             {
                printf("ERROR 67: cant find label %s\n", goto_label[i] );
                return(-1);
             }
           }

           return(0);

        }

        xp = line;
        while ( *xp != 0 )
        {
          if ( *xp == 10 || *xp == 13 )
		  {
		    *xp = 0;
            break;
		  }
          xp++;
        }

		stmt = line;
		cursor = line;

        if (line[0] == '/' && line[1] == '/' )
		{
          first_line = 1;  // force fgets in case line contains ;
          continue;
		}

        if (line[0] != '#')
          break;
	  }
	  else
	  {
		stmt = cursor;
		break;
	  }
    }

    cursor = stmt;

    type = NONE;
    sign = SIGNED;
    pointer = 0;
    gotname = 0;

    cursor = gettkn( token , cursor );


    if (token[0] == 0 || ( (short ) strcmp(token,";") ) == 0 )
    {
        continue; // no tokens on line - get next stmt 
    }


    if ( compile_and_go == 0 && compile_only == 0 )
    {
       printf("\n%3d: %04x: %s\n\n",lineno,codeoff,stmt);
       ngetchx();
    }


    //
    // decode a variable definition
    //

/////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////

    if ( (short) strcmp(token,"function") == 0 )
    {
       in_function = 1;
       in_function_arglist = 1;
       function_a6_offset = 8;
       num_function_args = 0;

       cursor = gettkn( token, cursor );
       if ( token[0] == 0 )
       {
          printf("ERROR: missing function name\n");
          return(81);
       }

       if ( num_functions == 0 )
       {
          branch_around_offset = codeoff+2;  // offset of fix
          opword = 0x6000;   // bra
          dispword = 0x0000; // to be fixed
          emit4();
          if (compile_listing == 1 )
             printf("bra    around functions\n");
       }

       strcpy( function_name[num_functions], token);
       function_code_offset[num_functions] = codeoff;
       num_functions++;

       
       opword = 0x4e56;   // link a6,#0
       dispword = 0x0000;
       emit4();
       if ( compile_listing == 1 )
          printf("link a6,#0\n");

       function_movem_offset = codeoff+2;  // for fixing regs

       if ( compile_listing == 1 )
          printf("function movem offset at 0x%04x\n",
             function_movem_offset);

       movem_a5_offset += 64;

       opword = 0x48ed;              // movem.l regs,d(a5)
       dispword = 0x0000;            // 0  regs to be fixed
       thirdword = movem_a5_offset;  // d
       emit6();

       if ( compile_listing == 1 )
          printf("movem.l regs,%d(a5)\n",movem_a5_offset);

       nextaddrreg = 1;
       nextdatareg = 2;

       movem_mask = 0;
    }
    else if ( (short) strcmp(token,"char") == 0
           || (short) strcmp(token,"long") == 0
           || (short) strcmp(token,"short") == 0
           || (short) strcmp(token,"reg") == 0
            )
    {
       if ( in_function_arglist == 1 )
       {
          data_or_reg = REG_VAR;
       }
       else if ( (short) strcmp(token,"reg") == 0 )
       {
          data_or_reg = REG_VAR;
          cursor = gettkn( token , cursor );

          if (token[0] == 0)
          {
             printf("ERROR: expect char/long/short/int after reg\n");
             return(3);
          }
       }
       else
       {
          data_or_reg = DATA_VAR;
       }


       if ( (short) strcmp(token,"char") == 0 )
       {
                type = CHAR;
       }
       else if ( (short) strcmp(token,"short") == 0 )
       {
          type = SHORT;
       }
       else if ( (short) strcmp(token,"long") == 0 )
       {
          type = LONG;
       }
       else
       {
          printf("syntax: data/reg char/short/long  [*] <varname>\n");
          return(4);
       }

       cursor = gettkn( token , cursor ); // get varname or *

       if (token[0] == 0)
       {
          printf("syntax: data/reg char/short/long  [*] <varname>\n");
          return(5);
       }

       if ( (short) strcmp(token,"*") == 0 )
       {
          pointer = 1;
          cursor = gettkn( token , cursor );

          if (token[0] == 0)
          {
             printf("syntax: data/reg char/short/long  [*] <varname>\n");
             return(6);
          }

       }

       if ( token[0] == '*' )
       {
          strcpy( varname, token+1 );
          pointer = 1;
       }
       else
       {
          strcpy( varname, token );
       }

       strcpy( basename, varname );

       if ( data_or_reg == DATA_VAR )
       {
          dim = 1;
          if ( strstr( basename, "[" ) != NULL )
          {
              dimp = NULL;
              xp = &basename[0];
              for (;;)
              {
                 if ( *xp == 0 )
                    break;
                 if ( *xp == '[' )
                 {
                    dimp = xp+1;
                    *xp = 0;
                    break;
                 }
                 xp++;
             }
             dim = 0;
             if ( dimp != NULL )
             {
                dim = atoi(dimp);
             }
          }
   
          length = type*dim;
          if (compile_listing == 1 )
             printf("// GLOBAL: %15s  len=%3d  dim=%4d  type=%d offset %4d\n",
                        varname,length,dim,type,a5_gbloff);
   
          strcpy( gblname[numgbls], basename );
          gbldim[ numgbls ] = dim;

          gbloffset[ numgbls ] = a5_gbloff;

          curp = a5_area + a5_gbloff;
          curshortp = (short *) curp;
          curlongp = (long *) curp;

          a5_gbloff += length;

          gbltype[ numgbls ] = type;

          if ( pointer == 1 )
              gblsize[ numgbls ] = 4;
          else
              gblsize[ numgbls ] = type;
          gblsign[ numgbls ] = sign;
          numgbls++;

          cursor = gettkn( token , cursor ); // get varname or *

          if (token[0] == 0)
          {
             printf("syntax: data/reg char/short/long  [*] <varname>[dim] ;/= initializers\n");
             return(7);
          }

          if ( (short) strcmp(token,"=") == 0 )  // initializer
          {
                for (;;)
                {
                   cursor = gettkn( token , cursor );

                   if (token[0] == 0)
                       break;
                   if ( token[0] == '{' )
                       continue;
                   if ( token[0] == '}' )
                       break;
                   if ( token[0] == ',' )
                       continue;
                   longval = strtol(token,NULL,0);
                   if ( type == 1 )
                      *curp++ = longval;
                   else if ( type == 2 )
                      *curshortp++ = longval;
                   else if ( type == 4 )
                      *curlongp++ = longval;
                   else
                   {
                      printf("error bad type for val %ld\n",longval);
                      return(70);
                   }
                }
          }

      }
      else  // data_or_reg == REG_VAR
      {
          if ( pointer == 1 ) // allocate addr reg
          {

              regno = nextaddrreg++;
              if ( regno > 4 )
              {
                  printf("ERROR: out of addr regs var=%s\n",
                  varname);
                  return(8);
              }
              strcpy(addrregname[regno],varname);
              addrregtosize[regno] =  type;

              if ( in_function_arglist == 1 )
              {
                  // movea.l  xxx(a6),ar
                  opword = 0x206e | (regno<<9);
                  dispword = function_a6_offset;

                  emit4();

                  if (compile_listing == 1 )
                     printf("movea.l  %d(a6),a%d\n",
                             function_a6_offset,regno);
                  function_a6_offset += 4;
              }


              if ( in_function == 1 )
              {
                  movem_mask |= (0x00000100 << regno);
              }
      
          }
          else
          {
              regno = nextdatareg++;
              if ( regno > 7 )
              {
                  printf("ERROR: out of data regs var=%s\n",
                  varname);
                  return(9);
              }
              strcpy(dataregname[regno],varname);
              dataregsize[regno] = type;
              datasign[regno] = sign;

              if ( in_function_arglist == 1 )
              {
                  // move.s  xxx(a6),dr

                  if ( type == 2  || type == 1 )
                    opword = 0x302e;
                  else if ( type == 4 )
                    opword = 0x202e;

                  opword |= (regno<<9); // dest reg
                  dispword = function_a6_offset;

                  emit4();

                  if (compile_listing == 1 )
                     printf("move.%d  %d(a6),d%d\n",
                        type,function_a6_offset,regno);

                  function_a6_offset += type;
              }

              if ( in_function == 1 )
              {
                  movem_mask |= (0x00000001 << regno);
              }
          }
      }    // else data_or_reg == REG_VAR

    }  // if data or reg
/////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////
    else if ( (short) strcmp(token,"for") == 0 )
    {
       brace_level++;
       brace_type[brace_level] = FOR;

       for_level++;
       numbreaks[for_level] = 0;

#ifdef CMDLINE_VERSION
#if 0
       printf("// LVL %d: ",for_level);
       printf("/ %s ",token);
       for (;;)
       {
          cursor = gettkn( token , cursor );

          if (token[0] == 0)
            break;
          printf("/ %s ",token);
       }
       printf("\n");
#endif
#endif
       cursor = stmt;
       cursor = gettkn( token , cursor );  // for

       cursor = gettkn( token , cursor );  // (
       if ( (short) strcmp(token,"(") != 0)
       {
          printf("syntax: ( missing\n");
          return(10);
       }

       cursor = gettkn( token , cursor ); // dreg

       for_dregno = getdregno( token, &for_dreglen );
       if ( for_dregno == 0 )
       {
          printf("syntax: %s not a valid left dreg\n",token);
          return(11);
       }

       cursor = gettkn( token, cursor );  // =
       if ( (short) strcmp(token,"=") != 0 )
       {
          printf("syntax: = missing got %s instead\n",token);
          return(11);
       }

       cursor = gettkn( token, cursor );  // 0

       if ( (short) strcmp(token,"0") != 0 )
       {
          printf("syntax: first 0 missing got %s instead\n",token);
          return(11);
       }

       cursor = gettkn( token, cursor );  // ;

       if ( (short) strcmp(token,";") != 0 )
       {
          printf("syntax: first ; missing got %s instead\n",token);
          return(11);
       }

       cursor = gettkn( token , cursor ); // dreg

       dregno = getdregno( token, &for_dreglen );
       if ( dregno != for_dregno )
       {
          printf("syntax: %s not a valid middle dreg\n",token);
          return(12);
       }

       cursor = gettkn( token , cursor ); // <

       if ( (short) strcmp(token,"<") != 0 )
       {
          printf("syntax: < missing: got %s instead\n",token);
          return(13);
       }

       cursor = gettkn( token, cursor );   // limit dreg/#limit

       limit = 0;
       limit_dregno = getdregno( token, &limit_dreglen );
       if ( limit_dregno == 0 )
       {
          limit = strtoul( token, NULL, 0 );
       }

       cursor = gettkn( token, cursor );   // ;

       if ( (short) strcmp(token,";") != 0)
       {
          printf("syntax: second ; missing: got %s instead\n",token);
          return(14);
       }

       cursor = gettkn( token , cursor ); // dreg

       dregno = getdregno( token, &for_dreglen );
       if ( dregno != for_dregno )
       {
          printf("syntax: %s not a valid right dreg\n",token);
          return(15);
       }

       cursor = gettkn( token , cursor );

       if ( (short) strcmp(token,"++") == 0 )   // ++
       {
           for_increment = 1;
       }
       else if ( (short) strcmp(token,"+=") == 0)   // += k
       {
           cursor = gettkn( token , cursor ); // constant
           for_increment = strtoul( token, NULL, 0 );
       }
       else
       {
          printf("syntax: ++ or += missing: got %s instead\n",token);
          return(16);
       }

       //
       // now emit the code
       // for the for stmt
       // using the above for_*
       //

       opword = 0x4240 | for_dregno;
       emit2();

       if (compile_listing == 1 )
          printf("clr.w    d%d\n",for_dregno);

       opword = 0x6000; // bra +4

       dispword = 0x0004;
       emit4();
       if (compile_listing == 1 )
          printf("bra +4\n");


       continue_offset[for_level] = codeoff;

       // addq.w %increment,dn
       opword = 0x5040 | ((for_increment & 0x7)<<9) | for_dregno;
     
       emit2();
       if (compile_listing == 1 )
          printf("addq.w #%ld,d%d\n",for_increment,for_dregno);

       if ( limit_dregno == 0 )   // cmpi.w #,dn
       {
           opword = 0x0c40 | for_dregno;
           dispword = limit;
           emit4();
           if (compile_listing == 1 )
              printf("cmpi.w %d,d%d\n",limit,for_dregno);
       }
       else   // cmp.w dl,dn
       {
           opword = 0xb040 | (for_dregno<<9) | limit_dregno;
           emit2();
           if (compile_listing == 1 )
              printf("cmp.w  d%d,d%d\n",limit_dregno,for_dregno);
       }

       opword = 0x6400; // bra   breakpt
       dispword = 0x0000;
       emit4();
       if (compile_listing == 1 )
          printf("bra   breakpt\n");


       //
       // code for a break statement at this for level
       // or lower if level
       //
       breakno = numbreaks[for_level];
       break_fix[for_level][breakno] = codeoff-2;
       numbreaks[for_level]++;

       //
       // end of break code
       //
       last_statement_type = FOR;
    }
/////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////
    else if ( (short) strcmp(token,"if") == 0 )
    {
       brace_level++;
       brace_type[brace_level] = IF;

#ifdef CMDLINE_VERSION
#if 0
       printf("// IF LVL %d: ",brace_level);

       printf("/ %s ",token);
       for (;;)
       {
          cursor = gettkn( token , cursor );
          if (token[0] == 0)
            break;
          printf("/ %s ",token);
       }
       printf("\n");
#endif
#endif

       cursor = stmt;

       cursor = gettkn( token , cursor );  // if

       cursor = gettkn( token , cursor );  // (
       if ( (short) strcmp(token,"(") != 0)
       {
          printf("syntax: if ( ref ... but ( missing after if\n");
          return(17);
       }

       cursor = gettkn( token , cursor );  // dreg


       errcode = parse_dataref();  // get token for compare op also
       if ( errcode != 0 )
       {
          printf("syntax: %s not a valid if variable reference\n",token);
          return(18);
       }


       if ( (short) strcmp(token,"==") == 0 )
       {
          comparison = EQ;
          else_mask = 0x6600; // bne
          strcpy(condition,"ne");
       }
       else if ( (short) strcmp(token,"!=") == 0 )
       {
          comparison = NE;
          else_mask = 0x6700; // beq
          strcpy(condition,"eq");
       }
       else if ( (short) strcmp(token,">=") == 0 )
       {
          comparison = GE;
          else_mask = 0x6500; // bcs
          strcpy(condition,"cs");
       }
       else if ( (short) strcmp(token,"<=") == 0 )
       {
          comparison = LE;
          else_mask = 0x6200; // bhi
          strcpy(condition,"hi");
       }
       else if ( (short) strcmp(token,">") == 0 )
       {
          comparison = GT;
          else_mask = 0x6300; // bls
          strcpy(condition,"ls");
       }
       else if ( (short) strcmp(token,"<") == 0 )
       {
          comparison = LT;
          else_mask = 0x6400; // bcc
          strcpy(condition,"cc");
       }
       else
       {
          printf("syntax: if ( dreg op #/dreg ) - bad op\n");
          return(19);
       }

       if_dreglen = datalen;

       if ( result_type == 5 ) // dreg
       {
         if_dregno = dregno;
       }
       else
       {
         dest_mask = 0x0200;  // move.s  xxx,d1

         // now compile the move to d1

         type = result_type;
         areg = aregno;
         dreg = dregno;
         con  = constant;
         dtype = datatype;
         dlen = datalen;
         ilen = indexlen;
         off  = offset;
       
         if ( dlen == 1 )
           len_mask = 0x1000;
         else if ( dlen == 2 )
           len_mask = 0x3000;
         else if ( dlen == 4 )
           len_mask = 0x2000;
         else
         {
           printf("internal error: dlen = %d\n",dlen);
           return(34);
         }

         switch( (short) type )
         {
            case 1:   // move.s (an)+,d1
               opword = 0x0018 | dest_mask | len_mask | areg;
               emit2();
               if (compile_listing == 1 )
                  printf("move.%d  (a%d)+,d%d\n",
                     dlen,areg,(dest_mask>>9));
               break;
       
            case 2:   // move.s (an),d1
               opword = 0x0010 | dest_mask | len_mask | areg;
               emit2();
               if (compile_listing == 1 )
                  printf("move.%d  (a%d),d%d\n",
                     dlen,areg,(dest_mask>>9));
               break;
       
            case 3:   // move.s d(an),d1
               opword = 0x0028 | dest_mask | len_mask | areg;
               shortcon = con;
               dispword = shortcon*dlen;
               emit4();
               if (compile_listing == 1 )
               {
                      printf("move.%d  %d(a%d),d%d\n",
                             dlen,dispword,areg,(dest_mask>>9));
               }
               break;
       
            case 4:   // move.s 0(an,di),d1

               opword = 0x0030 | areg | dest_mask | len_mask;
               dispword = 0x0000;  // D/Aiii W/L000 0000 0000
               if ( ilen == 2 )
               {
                 dispword = 0x0000 | ( dreg << 12 );
               }
               else if ( ilen == 4 )
               {
                 dispword = 0x0800 | ( dreg << 12 );
               }
               else
               {
                 printf("internal error: bad ilen = %d\n",ilen);
                 return(35);
               }
               emit4();
               if (compile_listing == 1 )
               {
                   printf("move.%d  0(a%d,d%d.%d),d%d\n",
                          dlen,areg,dreg,ilen,(dest_mask>>9));
               }
               break;
       
            case 5:   // move.s dn,d1
               opword = 0x0000 | dest_mask | len_mask | dreg;
               emit2();
               if (compile_listing == 1 )
               {
                   printf("move.%d  d%d,d%d\n",
                          dlen,dreg,(dest_mask>>9));
               }
               break;
       
            case 6:   // move.l an,d1
               opword = 0x2008 | dest_mask | areg ;
               emit2();
               if (compile_listing == 1 )
               {
                   printf("move.4  a%d,d%d\n",
                          areg,(dest_mask>>9));
               }
               break;
       
            case 7:   // move.s d(a5),d1
               opword = 0x002d | dest_mask | len_mask;
               dispword = off;
               emit4();
               if (compile_listing == 1 )
               {
                   printf("move.%d  %d(a5),d%d\n",
                          dlen,dispword,(dest_mask>>9));
               }
               break;
       
            case 8:   // move.s d(a5),d1
               opword = 0x002d | dest_mask | len_mask;
               shortcon = con;
               dispword = off + (shortcon*dlen);
               emit4();
               if (compile_listing == 1 )
               {
                   printf("move.%d  %d(a5),d%d\n",
                          dlen,dispword,(dest_mask>>9));
               }
               break;
       
            case 9:   // move.s d(an,di),d1

               opword = 0x0030 | dest_mask | areg | len_mask;

               if ( ilen == 2 )
               {
                 // D/Aiii W/L000 dddd dddd
                 dispword = 0x0000 | ( dreg << 12 ) | off;
               }
               else if ( ilen == 4 )
               {
                 // D/Aiii W/L000 dddd dddd
                 dispword = 0x0800 | ( dreg << 12 ) | off; 
               }
               else
               {
                 printf("internal error: bad ilen = %d\n",ilen);
                 return(36);
               }
               emit4();
               if (compile_listing == 1 )
               {
                   printf("move.%d  %d(a%d,d%d.%d),d%d\n",
                          dlen,off,areg,dreg,ilen,(dest_mask>>9));
               }
               break;
       
            case 11:
               // lea d(a5),a0
               opword = 0x41ed;
               dispword = off;
               emit4();
               if (compile_listing == 1 )
                  printf("lea     %d(a5),a0\n", off);

               // move a0,dr
               opword = 0x2008 | dest_mask;
               emit2();
               if (compile_listing == 1 )
                  printf("move a0,d%d\n", (dest_mask>>9) );
               break;
       
            default:
               printf("internal error: 3 bad rtype = %d\n",type);
               break;
         }

         if_dregno = 1;
       }

       // now compile the compare

       if ( if_dreglen == 4 )      // 4 - long
           len_mask = 0x0080;
       else if ( if_dreglen == 2 ) // 2 - short
           len_mask = 0x0040;
       else                  // 1 - byte
           len_mask = 0x0000;


       cursor = gettkn( token , cursor );   //  src reg or constant

       compare_dregno = getdregno( token, &compare_dreglen );
       if ( compare_dregno == 0 )
       {
          //
          // IMMEDIATE  dn += #imm
          //

          if ( token[0] == '\'' )
            immediate = token[1];
          else
            immediate = strtoul( token, NULL, 0 );

          opword =  0x0c00 | len_mask | if_dregno;

          if ( if_dreglen == 2 || if_dreglen == 1 )
          {
              dispword = immediate;
              emit4();
          }
          else if ( dlen == 4 )
          {
              dispword = immediate>>16;
              thirdword = immediate & 0xffff;
              emit6();
          }
          if (compile_listing == 1 )
             printf("cmp.%d   #%ld,d%d\n",
                   if_dreglen,immediate,if_dregno);
       }
       else
       {
          opword = 0xb000 | (if_dregno<<9) | compare_dregno | len_mask;
          emit2();
          if (compile_listing == 1 )
             printf("cmp.%d   d%d,d%d\n",
                   if_dreglen,compare_dregno,if_dregno);
       }

       opword = else_mask;
       dispword = 0x0000;

       if_fix_offset = codeoff+2;

       emit4();
       if (compile_listing == 1 )
          printf("b%s     else_pt\n",condition);


       brace_if_fix[brace_level] = if_fix_offset;

       last_statement_type = IF;
    }
/////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////
    else if ( (short) strcmp(token,"goto") == 0 )
    {
       cursor = gettkn( token, cursor );
       if ( token[0] == 0 )
       {
          printf("error 66: goto statement need label\n");
          return(-1);
       }

       opword = 0x6000;
       dispword = 0x0000;
       emit4();
       if (compile_listing == 1 )
          printf("bra    %s\n",token);

       goto_fix_offset[numgotos] = codeoff-2;
       strcpy( goto_label[numgotos], token);
       strcat( goto_label[numgotos],":");
       numgotos++;
    }
/////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////
    else if ( *(token+strlen(token)-1) == ':' )
    {
       strcpy( label[ numlabels ] , token );
       labeloff[ numlabels ] = codeoff;
       numlabels++;
    }
/////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////
    else if ( (short) strcmp(token,"break") == 0 )
    {
       // code for a break statement at this for level
       // or lower if level
       //
       opword = 0x6000;
       dispword = 0x0000;
       emit4();
       if (compile_listing == 1 )
          printf("bra     break_pt\n");

       breakno = numbreaks[for_level];
       break_fix[for_level][breakno] = codeoff-2;
       numbreaks[for_level]++;

       //
       // end of break code
       //
    }
/////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////
    else if ( (short) strcmp(token,"continue") == 0 )
    {
          //
          // emit code for continue at this for_level
          //

          continue_off = continue_offset[for_level]-(codeoff+2);

          opword = 0x6000;
          dispword = continue_off;
          emit4();
          if (compile_listing == 1 )
             printf("bra   continue_pt (%d)\n",continue_off);
    }
/////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////
    else if ( (short) strcmp(token,"return") == 0 )
    {
        // use movem_mask

       cursor = gettkn( token, cursor );
       if ( token[0] == 0 )
       {
           printf("ERROR: return needs argument\n");
           return(83);
       }

       dregno = getdregno(token,&set_dreglen);
       if ( dregno == 0 )
       {
          aregno = getaregno(token,&set_areglen);
          if ( aregno == 0 )
          {
             printf("ERROR: syntax: return areg/dreg;\n");
             return(87);
          }
          // movea.l an,a0
          // move.l  a0,d0

          opword = 0x2048 | aregno;  // movea.l an,a0
          emit2();
          if ( compile_listing == 1 )
              printf("movea.l a%d,a0\n",aregno);

          opword = 0x2008;           // move.l a0,d0
          emit2();
          if ( compile_listing == 1 )
              printf("move.l  a0,d0\n");

       }
       else
       {
          // move.l  dn,d0
          // movea.l d0,a0

          opword = 0x2000 | dregno;  // move.l dn,d0
          emit2();
          if ( compile_listing == 1 )
              printf("move.l  d%d,d0\n",dregno);


          opword = 0x2040;           // movea.l d0,a0
          emit2();
          if ( compile_listing == 1 )
              printf("movea.l d0,a0\n");



       }

       opword = 0x4ced;              // movem.l d(a5),regs
       dispword = movem_mask;        // regs to save
       thirdword = movem_a5_offset;  // d
       emit6();

       if ( compile_listing == 1 )
          printf("movem.l %d(a5),regs:0x%04x\n",
             movem_a5_offset,movem_mask);

       opword = 0x4e5e;  // unlk a6
       emit2();

       if ( compile_listing == 1 )
          printf("unlk a6\n",movem_a5_offset);

       opword = 0x4e75;  // rts
       emit2();

       if ( compile_listing == 1 )
          printf("rts\n");

    }
/////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////
    else if ( (short) strcmp(token,"{") == 0 )
    {
       if ( in_function_arglist == 1 )
       {
          in_function_arglist = 0;
       }
    }
/////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////
    else if ( (short) strcmp(token,"}") == 0 )
    {
        //
        // use brace_type[brace_level]
        //
        // for ifs, fix branch at
        // brace_if_fix[brace_level]
        // to branch to the current offset
        //

        if ( brace_level == 0 && in_function == 1 ) // end of function
        {
           // use movem_mask

           opword = 0x48ed;              // movem.l regs,d(a5)
           dispword = movem_mask;        // regs to save
           thirdword = movem_a5_offset;  // d
           emit6();
    
           if ( compile_listing == 1 )
              printf("movem.l %d(a5),regs\n",movem_a5_offset);
    
           opword = 0x4e5e;  // unlk a6
           emit2();
    
           if ( compile_listing == 1 )
              printf("unlk a6\n",movem_a5_offset);
    
           opword = 0x4e75;  // rts
           emit2();

           if ( compile_listing == 1 )
              printf("rts\n");

           cp = codeptr + branch_around_offset;
           codep = (unsigned short *) cp;
           *codep = (unsigned short ) (codeoff-branch_around_offset);

           if ( compile_listing == 1 )
              printf("branch_around at 0x%04x fixed to 0x%04x\n",
                branch_around_offset,(codeoff-branch_around_offset) );

           cp = codeptr+function_movem_offset;
           codep = (unsigned short *) cp;
           *codep = movem_mask;

           if ( compile_listing == 1 )
              printf("movem mask fix at offset 0x%04x to 0x%04x\n",
                  function_movem_offset,movem_mask);

           in_function = 0;
           in_function_arglist = 0;
           nextaddrreg = 1;
           nextdatareg = 2;
        }
        else if ( brace_type[brace_level] == IF )
        {
           fix_offset = brace_if_fix[brace_level];
           if (compile_listing == 1 )
              printf("code[0x%04x] = 0x%04x;\n",
                   fix_offset,codeoff-fix_offset);

           cp = codeptr+fix_offset;
           codep = (unsigned short *) cp;
           *codep = codeoff-fix_offset;
           brace_level--;
        }
        else if ( brace_type[brace_level] == FOR )
        {
           //
           // emit code for continue at this for_level
           //

           continue_off = continue_offset[for_level]-(codeoff+2);

           opword = 0x6000;
           dispword = continue_off;
           emit4();

           if (compile_listing == 1 )
              printf("bra   continue_pt (%d)\n",continue_off);

           breaks = numbreaks[for_level];
           for (i=0;i<breaks;i++)
           {
              fix_offset = break_fix[for_level][i];
              if (compile_listing == 1 )
                 printf("code[0x%04x] = 0x%04x;\n",
                   fix_offset,codeoff-fix_offset);

              cp = codeptr+fix_offset;
              codep = (unsigned short *) cp;
              *codep = codeoff-fix_offset;

           }
	   for_level--;
           brace_level--;
        }
        //
    }
/////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////
    else if (token[0] == 0 )
    {
    }
/////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////
    else if ( strstr(token,"(") != NULL
              || ((short) strcmp(token,"set")) == 0 )
    {

       set_flag = 0;

       if ( (short) strcmp(token,"set") == 0 )
       {
          set_flag = 1;

          //
          // set regvar = function( arg, arg, ...
          //

          set_aregno = 0;
          cursor = gettkn( token, cursor );
          set_dregno = getdregno( token, &set_dreglen );
          if ( set_dregno == 0 )
          {
              set_aregno = getaregno( token, &set_areglen );
              if ( set_aregno == 0 )
              {
                 printf("syntax:  set a/dreg = function( ...\n");
                 return(20);
              }
          }
          cursor = gettkn( token, cursor );
          if ( (short) strcmp(token,"=") != 0 )
          {
            printf("syntax:  set a/dreg = function( ...\n");
            return(21);
          }
          call_cursor = cursor;  // after =
       }
       else
       {
          call_cursor = stmt;
       }

       cursor = stmt;

#ifdef CMDLINE_VERSION
#if 0
       for (;;)
       {
          cursor = gettkn( token , cursor );

          if (token[0] == 0)
            break;
          printf("/ %s ",token);
       }
       printf("\n");
#endif
#endif

       cursor = call_cursor;

       cursor = gettkn(token,cursor);  //  function(
       
       xp = token;
       while ( *xp != 0 )
       {
         if (*xp == '(')
         {
            *xp = 0;
            break;
         }
         xp++;
       }

       strcpy( functionname, token );

       numargs = 0;

       for (;;)
       {  
          cursor = gettkn(token,cursor);  // first arg if any
          if ( (short) strcmp(token,")") == 0 )
              break;

          errcode = parse_dataref();
          if ( errcode != 0 )
          {
            printf("error %d from parse_dataref\n",errcode);
            return(50);
          }

          arg_type[numargs] = result_type;
          arg_aregno[numargs] = aregno;
          arg_dregno[numargs] = dregno;
          arg_constant[numargs] = constant;
          arg_offset[numargs]   = offset;
          arg_datatype[numargs] = datatype;
          arg_datalen[numargs]  = datalen;
          arg_indexlen[numargs]  = indexlen;
          numargs++;

          if ( (short) strcmp(token,")") == 0 )
            break;

          if ( (short) strcmp(token,",") != 0 )
          {
              printf("expected , got %s instead\n",token);
              return(22);
          }

       }

       //
       // now emit code for function call
       // using above in reverse order
       //

       total_len = 0;

       for (arg = numargs-1;arg>=0;arg--)
       {

         type = arg_type[arg];
         areg = arg_aregno[arg];
         dreg = arg_dregno[arg];
         con  = arg_constant[arg];
         dtype = arg_datatype[arg];
         dlen = arg_datalen[arg];
         ilen = arg_indexlen[arg];
         off  = arg_offset[arg];
       
         if ( dlen == 1 )
           len_mask = 0x3000;  // push as word for stack
         else if ( dlen == 2 )
           len_mask = 0x3000;
         else if ( dlen == 4 )
           len_mask = 0x2000;
         else
         {
           printf("internal error: dlen = %d\n",dlen);
           return(23);
         }
//
// result types:
//
//            1   *areg++
//            2   *areg
//            3   areg[constant]
//            4   areg[dreg]
//            5   dreg
//            6   areg
//            7   data
//            8   data[constant]
//            9   data[dreg]
//            10  constant
//            11  &data
//            12  "literal string"
//
       
         switch( (short ) type )
         {
            case 1:   // move.s (an)+,-(a7)
               opword = 0x0f18 | len_mask | areg;
               emit2();
               if (compile_listing == 1 )
                  printf("move.%d  (a%d)+,-(a7)\n",dlen,areg);
               break;
       
            case 2:   // move.s (an),-(a7)
               opword = 0x0f10 | len_mask | areg;
               emit2();
               if (compile_listing == 1 )
                  printf("move.%d  (a%d),-(a7)\n",dlen,areg);
               break;
       
            case 3:   // move.s d(an),-(a7)
               opword = 0x0f28 | len_mask | areg;
               shortcon = con;
               dispword = shortcon*dlen;
               emit4();
               if (compile_listing == 1 )
                  printf("move.%d  %d(a%d),-(a7)\n",dlen,dispword,areg);
               break;
       
            case 4:   // move.s 0(an,di),-(a7)
               opword = 0x0f30 | areg | len_mask;
               dispword = 0x0000;  // D/Aiii W/L000 0000 0000
               if ( ilen == 2 )
               {
                 dispword = 0x0000 | ( dreg << 12 );
               }
               else if ( ilen == 4 )
               {
                 dispword = 0x0800 | ( dreg << 12 );
               }
               else
               {
                 printf("internal error: bad ilen = %d\n",ilen);
                 return(24);
               }
               emit4();
               if (compile_listing == 1 )
                  printf("move.%d  0(a%d,d%d.%d),-(a7)\n",
                        dlen,areg,dreg,ilen);
               break;
       
            case 5:   // move.s dn,-(a7)
               opword = 0x0f00 | len_mask | dreg;
               emit2();
               if (compile_listing == 1 )
                  printf("move.%d  d%d,-(a7)\n",dlen,dreg);
               break;
       
            case 6:   // move.l an,-(a7)
               opword = 0x2f08 | areg ;
               emit2();
               if (compile_listing == 1 )
                  printf("move.4  a%d,-(a7)\n",areg);
               break;
       
            case 7:   // move.s d(a5),-(a7)
               opword = 0x0f2d | len_mask;
               dispword = off;
               emit4();
               if (compile_listing == 1 )
                  printf("move.%d  %d(a5),-(a7)\n",dlen,dispword);
               break;
       
            case 8:   // move.s d(a5),-(a7)
               opword = 0x0f2d | len_mask;
               shortcon = con;
               dispword = off + (shortcon*dlen);
               emit4();
               if (compile_listing == 1 )
                  printf("move.%d  %d(a5),-(a7)\n",dlen,dispword);
               break;
       
            case 9:   // move.s d(an,di),-(a7)
               opword = 0x0f30 | areg | len_mask;
               if ( ilen == 2 )
               {
                 // D/Aiii W/L000 dddd dddd
                 dispword = 0x0000 | ( dreg << 12 ) | off;
               }
               else if ( ilen == 4 )
               {
                 // D/Aiii W/L000 dddd dddd
                 dispword = 0x0800 | ( dreg << 12 ) | off; 
               }
               else
               {
                 printf("internal error: bad ilen = %d\n",ilen);
                 return(25);
               }
               emit4();
               if (compile_listing == 1 )
                  printf("move.%d  %d(a%d,d%d.%d),-(a7)\n",
                        dlen,off,areg,dreg,ilen);
               break;
       
            case 10:   // move.l #imm,-(a7)
               opword = 0x0f3c | len_mask;
               dispword = con;

               if ( dlen == 2 )
               {
                   emit4();
               }
               else
               {
                   dispword = con>>16;
                   thirdword = con & 0xffff;
                   emit6();
               }
               if (compile_listing == 1 )
                   printf("move.%d  #%ld,-(a7)\n",dlen,con);

               break;
       
            case 11:   // pea  d(a5)
            case 12:
               opword = 0x486d;
               dispword = off;
               emit4();
               if (compile_listing == 1 )
                  printf("pea     %d(a5)\n",dispword);
               break;
       
            default:
               printf("internal error: 2 bad rtype = %d\n",type);
               break;
         }

         if ( dlen == 1 )
           total_len += 2;
         else
           total_len += dlen;

       }

       for (i=0;i<num_functions;i++)
       {
          if ( (short) strcmp( functionname,
                               function_name[i] )
               == 0
             )
          {
            opword = 0x6100;  // bsr
            dispword = function_code_offset[i]-codeoff-2;
            emit4();
            if (compile_listing == 1 )
               printf("bsr   %d\n",dispword);
            goto do_post_call;
          }
       }


       opword = 0x48ed;
       dispword = 0x0e0e;
       thirdword = 0x0040;
       emit6();
       if (compile_listing == 1 )
          printf("movem.l a3-a1-d3-d1,64(a5)\n");

       if (    (short) strcmp(functionname,"printf") == 0
            || (short) strcmp(functionname,"pf") == 0
          )
       {
           opword = 0x206d;
           dispword = 0x0000;
           emit4();
           if (compile_listing == 1 )
              printf("movea.l 0(a5),a0  printf\n");

           opword = 0x202d;
           dispword = 0x0004;
           emit4();
           if (compile_listing == 1 )
              printf("move.l  4(a5),d0  putchar\n");

           opword = 0x2f00;
           emit2();
           if (compile_listing == 1 )
              printf("move.l  d0,-(a7)\n");

           opword = 0x4e90;
           emit2();
           if (compile_listing == 1 )
              printf("jsr (a0)\n");

           opword = 0x588f;
           emit2();
           if (compile_listing == 1 )
              printf("addq.l #4,a7\n");
       }
       else if (  (short) strcmp(functionname,"gets") == 0 )
       {
           opword = 0x206d;
           dispword = 0x0008;
           emit4();
           if (compile_listing == 1 )
              printf("movea.l 8(a5),a0  gets\n");

           opword = 0x4e90;
           emit2();
           if (compile_listing == 1 )
              printf("jsr (a0)\n");
       }
       else if (  (short) strcmp(functionname,"fopen") == 0 )
       {
           opword = 0x206d;
           dispword = 0x000c;
           emit4();
           if (compile_listing == 1 )
              printf("movea.l 12(a5),a0  fopen\n");

           opword = 0x4e90;
           emit2();
           if (compile_listing == 1 )
              printf("jsr (a0)\n");
       }
       else if (  (short) strcmp(functionname,"fgetc") == 0 )
       {
           opword = 0x206d;
           dispword = 0x0010;
           emit4();
           if (compile_listing == 1 )
              printf("movea.l 16(a5),a0  fgetc\n");

           opword = 0x4e90;
           emit2();
           if (compile_listing == 1 )
              printf("jsr (a0)\n");
       }
       else if (  (short) strcmp(functionname,"fgets") == 0 )
       {
           opword = 0x206d;
           dispword = 0x0014;
           emit4();
           if (compile_listing == 1 )
              printf("movea.l 20(a5),a0  fgets\n");

           opword = 0x202d;
           dispword = 0x0010;
           emit4();
           if (compile_listing == 1 )
              printf("move.l  16(a5),d0  fgetc\n");

           opword = 0x2f00;
           emit2();
           if (compile_listing == 1 )
              printf("move.l  d0,-(a7)\n");

           opword = 0x4e90;
           emit2();
           if (compile_listing == 1 )
              printf("jsr (a0)\n");

           opword = 0x588f;
           emit2();
           if (compile_listing == 1 )
              printf("addq.l #4,a7\n");
       }
       else if (  (short) strcmp(functionname,"fprintf") == 0 )
       {
           opword = 0x206d;
           dispword = 0x001c;
           emit4();
           if (compile_listing == 1 )
              printf("movea.l 28(a5),a0  fprintf\n");

           opword = 0x202d;
           dispword = 0x0020;
           emit4();
           if (compile_listing == 1 )
              printf("move.l  32(a5),d0  fputc\n");

           opword = 0x2f00;
           emit2();
           if (compile_listing == 1 )
              printf("move.l  d0,-(a7)\n");

           opword = 0x4e90;
           emit2();
           if (compile_listing == 1 )
              printf("jsr (a0)\n");

           opword = 0x588f;
           emit2();
           if (compile_listing == 1 )
              printf("addq.l #4,a7\n");
       }
       else if (  (short) strcmp(functionname,"fclose") == 0 )
       {
           opword = 0x206d;
           dispword = 0x0018;
           emit4();
           if (compile_listing == 1 )
              printf("movea.l 24(a5),a0  fclose\n");

           opword = 0x4e90;
           emit2();
           if (compile_listing == 1 )
              printf("jsr (a0)\n");
       }
       else if (  (short) strcmp(functionname,"atoi") == 0 )
       {
           opword = 0x206d;
           dispword = 0x0024;
           emit4();
           if (compile_listing == 1 )
              printf("movea.l 36(a5),a0  atoi\n");

           opword = 0x4e90;
           emit2();
           if (compile_listing == 1 )
              printf("jsr (a0)\n");
       }
       else if (  (short) strcmp(functionname,"atol") == 0 )
       {
           opword = 0x206d;
           dispword = 0x0028;
           emit4();
           if (compile_listing == 1 )
              printf("movea.l 40(a5),a0  atol\n");

           opword = 0x4e90;
           emit2();
           if (compile_listing == 1 )
              printf("jsr (a0)\n");
       }
       else if (  (short) strcmp(functionname,"strtol") == 0 )
       {
           opword = 0x206d;
           dispword = 0x002c;
           emit4();
           if (compile_listing == 1 )
              printf("movea.l 44(a5),a0  strtol\n");

           opword = 0x4e90;
           emit2();
           if (compile_listing == 1 )
              printf("jsr (a0)\n");
       }
       else if (  (short) strcmp(functionname,"strtoul") == 0 )
       {
           opword = 0x206d;
           dispword = 0x0030;
           emit4();
           if (compile_listing == 1 )
              printf("movea.l 48(a5),a0  strtoul\n");

           opword = 0x4e90;
           emit2();
           if (compile_listing == 1 )
              printf("jsr (a0)\n");
       }
       else if (  (short) memcmp(functionname,"rom",3) == 0 )
       {
           romcall = strtoul( functionname+3, NULL, 16 );

           if ( romcall == 0 )
           {
              printf("error: function %s not a valid rom call\n",functionname);
              return(26);
           }

           opword = 0x2078;
           dispword = 0x00c8;
           emit4();
           if (compile_listing == 1 )
              printf("movea.l $c8,a0\n");
           opword = 0x2068;
           dispword = romcall*4;
           emit4();
           if (compile_listing == 1 )
              printf("movea.l 0x%x*4(a0),a0\n",romcall);
           opword = 0x4e90;
           emit2();
           if (compile_listing == 1 )
              printf("jsr    (a0)\n");
       }
       else
       {
           romcall = getromcall( functionname );
           if ( romcall == 0 )
           {
              printf("error: function %s not a valid rom call\n",functionname);
              return(26);
           }

           opword = 0x2078;
           dispword = 0x00c8;
           emit4();
           if (compile_listing == 1 )
              printf("movea.l $c8,a0\n");
           opword = 0x2068;
           dispword = romcall*4;
           emit4();
           if (compile_listing == 1 )
              printf("movea.l 0x%x*4(a0),a0\n",romcall);
           opword = 0x4e90;
           emit2();
           if (compile_listing == 1 )
              printf("jsr    (a0)\n");
       }

       opword = 0x4ced;
       dispword = 0x0e0e;
       thirdword = 0x0040;
       emit6();
       if (compile_listing == 1 )
          printf("movem.l 64(a5),a3-a1-d3-d1\n");

do_post_call:

       opword = 0x4fef;
       dispword = total_len;
       emit4();
       if (compile_listing == 1 )
          printf("lea     %d(a7),a7\n",total_len);

       if ( set_flag == 1 )
       {
            if ( set_dregno != 0 )
            {
               // move.s  d0,dn

               dlen = set_dreglen;
               if ( dlen == 4 )      // - long
                   op_mask = 0x2000;
               else if ( dlen == 2 ) // - short
                   op_mask = 0x3000;
               else                  // - byte
                   op_mask = 0x1000;

               opword = op_mask | (set_dregno<<9);
               emit2();
               if (compile_listing == 1 )
                  printf("move.%d  d0,d%d\n",set_dreglen,set_dregno);
            }
            else // set_aregno != 0
            {
               // movea.l  a0,an
               opword = 0x2048 | (set_aregno<<9);
               emit2();
               if (compile_listing == 1 )
                  printf("move.l   a0,a%d\n",set_aregno);
            }
       }
    }
/////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////
    else  // must be assignment
    {

#ifdef CMDLINE_VERSION
#if 0
       printf("/ %s ",token);
       for (;;)
       {
          cursor = gettkn( token , cursor );

          if (token[0] == 0)
            break;
          printf("/ %s ",token);
       }
       printf("\n");
#endif
#endif

       cursor = stmt;

       cursor = gettkn( token , cursor ); // dataref:

       errcode = parse_dataref();
       if ( errcode != 0 )
       {
          printf("error %d from parse_dataref\n",errcode);
          return(51);
       }


       //
       //  yields:
       //  
       //   result_type:
       //   aregno
       //   dregno
       //   constant
       //   offset
       //   datatype
       //   datalen
       //   indexlen
       //


       rslt_type = result_type;

       rslt_aregno = aregno;
       rslt_dregno = dregno;
       rslt_constant = constant;
       rslt_offset   = offset;
       rslt_datatype = datatype;
       rslt_datalen  = datalen;
       rslt_indexlen = indexlen;

       if ( (short) strcmp(token,"=") != 0 )
       {
          operator = NOOP;

          if (      (short) strcmp(token,"+=") == 0 )
          {
               operator = ADD;
          }
          else if ( (short) strcmp(token,"-=") == 0 )
          {
               operator = SUB;
          }
          else if ( (short) strcmp(token,"&=") == 0 )
          {
               operator = AND;
          }
          else if ( (short) strcmp(token,"|=") == 0 )
          {
               operator = OR;
          }
          else if ( (short) strcmp(token,"^=") == 0 )
          {
               operator = XOR;
          }
          else if ( (short) strcmp(token,">>") == 0 )
          {
               operator = SHR;
          }
          else if ( (short) strcmp(token,"<<") == 0 )
          {
               operator = SHL;
          }
          else if ( (short) strcmp(token,"<<<") == 0 )
          {
               operator = ROL;
          }
          else if ( (short) strcmp(token,">>>") == 0 )
          {
               operator = ROR;
          }
          else
          {
               printf("syntax: <rslt> = ... but = missing\n");
               return(27);
          }
          //
          // at this point operator is set
          // parse unary operator and emit code
          //
        
          if ( operator == ADD || operator == SUB )
          {
              if (rslt_type != 6 && rslt_type != 5 )
              {
                  printf("rslt must be d or a reg\n");
                  return(28);
              }
          }
          else if ( rslt_type != 5 )
          {
             printf("rslt type must be d reg\n");
             return(29);
          } 

          if ( rslt_datalen == 4 )  // long
             len_mask = 0x0080;
          else if ( rslt_datalen == 2 ) // short
             len_mask = 0x0040;
          else
             len_mask = 0x0000; // byte

          if ( rslt_type == 5 )   // dn op= ...
          {
             dest_mask = rslt_dregno;
   
             switch ( operator )
             {
                case ADD:
                  op_mask = 0xd100;
                  imm_mask = 0x0600;
                  strcpy(opname,"add");
                  break;
                case SUB:
                  op_mask = 0x9100;
                  imm_mask = 0x0400;
                  strcpy(opname,"sub");
                  break;
                case OR:
                  op_mask = 0x8100;
                  imm_mask = 0x0000;
                  strcpy(opname,"or ");
                  break;
                case AND:
                  op_mask = 0xc100;
                  imm_mask = 0x0200;
                  strcpy(opname,"and");
                  break;
                case XOR:
                  op_mask = 0xb100;
                  imm_mask = 0x0a00;
                  strcpy(opname,"eor");
                  break;
                case SHR:
                  op_mask =  0xe028;
                  imm_mask = 0xe008;
                  strcpy(opname,"lsr");
                  break;
                case SHL:
                  op_mask =  0xe128;
                  imm_mask = 0xe108;
                  strcpy(opname,"lsl");
                  break;
                case ROR:
                  op_mask =  0xe038;
                  imm_mask = 0xe018;
                  strcpy(opname,"ror");
                  break;
                case ROL:
                  op_mask =  0xe138;
                  imm_mask = 0xe118;
                  strcpy(opname,"rol");
                  break;
                default:
                  strcpy(opname,"none");
                  break;
             }

             cursor = gettkn( token , cursor );   //  src reg or constant

             src_regno = getdregno( token, &xreglen );
             if ( src_regno == 0 )
             {
                //
                // IMMEDIATE  dn += #imm
                //
                immediate = strtoul( token, NULL, 0 );
                if (   operator == ADD
                    && immediate >= 1
                    && immediate <= 8
                   )                   // addq #,dn
                {
                   immediate &= 0x7;
                   imm_mask = immediate << 9;
                   opword = 0x5000 | len_mask | dest_mask | imm_mask;
                   emit2();
                   if (compile_listing == 1 )
                      printf("addq.%d  %ld,d%d\n",
                           rslt_datalen,immediate,rslt_dregno);
                }
                else if ( operator == SUB
                       && immediate >= 1
                       && immediate <= 8
                        )                   // subq #,dn
                {
                   immediate &= 0x7;
                   imm_mask = immediate << 9;
                   opword = 0x5100 | len_mask | dest_mask | imm_mask;
                   emit2();
                   if (compile_listing == 1 )
                      printf("subq.%d  %ld,d%d\n",
                           rslt_datalen,immediate,rslt_dregno);
                }
                else if (   operator == SHR
                         || operator == SHL
                         || operator == ROR
                         || operator == ROL
                        )
                {
				   if (immediate > 8)
				   {
					  // immediate bits of 0 mean shift of 8
                      opword = len_mask | dest_mask | imm_mask;
                      emit2();
                      if (compile_listing == 1 )
                        printf("%s.%d   8,d%d\n",
                         opname,rslt_datalen,rslt_dregno);
				      immediate -= 8;
				   }

                   immediate &= 0x7;
                   opword = (immediate<<9) | len_mask | dest_mask | imm_mask;
                   emit2();
                   if (  immediate == 0 )
                     immediate = 8;
                   if (compile_listing == 1 )
                      printf("%s.%d   %ld,d%d\n",
                              opname,rslt_datalen,immediate,rslt_dregno);
                }
                else if ( rslt_datalen == 4 )
                {
                   {
                      opword = dest_mask | len_mask | imm_mask;
                      dispword = immediate>>16;
                      thirdword = immediate&0xffff;
                      emit6();
                      if (compile_listing == 1 )
                         printf("%s.%d   0x%lx,d%d\n",
                           opname,rslt_datalen,immediate,rslt_dregno);
                   }
                }
                else
                {
                   opword = dest_mask | len_mask | imm_mask;
                   dispword = immediate;
                   emit4();
                   if (compile_listing == 1 )
                      printf("%s.%d   0x%lx,d%d\n",
                              opname,rslt_datalen,immediate,rslt_dregno);
                }
             }
             else   // dn += dn
             {
               if ( operator != XOR )
               {
                 opword = (op_mask & 0xf000) | src_regno | (rslt_dregno<<9) | len_mask;
               }
               else
               {
                 opword = (src_regno<<9) | dest_mask | len_mask | op_mask;
               }
               emit2();
               if (compile_listing == 1 )
                  printf("%s.%d   d%d,d%d\n",
                         opname,rslt_datalen,src_regno,rslt_dregno);

             }
          }
          else  if ( rslt_type == 6 )  // an += #imm
          {
             cursor = gettkn( token , cursor );   //  imm constant
             immediate = strtoul( token, NULL, 0 );
             opword = 0x41e8 | (rslt_aregno<<9) | rslt_aregno;

             dispword = immediate;
             emit4();
             if (compile_listing == 1 )
                printf("lea     %ld(a%d),a%d\n",
                        immediate,rslt_aregno,rslt_aregno);
          }

          goto done_assign;
       }
       //
       // "=" found ==> assignment
       //
       cursor = gettkn( token , cursor ); // token after =

       if ( token[0] == 0 )
       {
          printf("syntax: <rslt> = ... but ... missing\n");
          return(30);
       }

       errcode = parse_dataref();  // token after =
       if ( errcode != 0 )
       {
          printf("error %d from parse_dataref\n",errcode);
          return(52);
       }


       left_type = result_type;
       left_aregno = aregno;
       left_dregno = dregno;
       left_constant = constant;
       left_offset   = offset;
       left_datatype = datatype;
       left_datalen  = datalen;
       left_indexlen = indexlen;
#if 0
       if ( token[0] == 0 )
       {
          printf("syntax: <rslt> = <lt> <op> <rt> but <op> missing\n");
          return(31);
       }
#endif

       if ( (short ) strcmp(token,"^") == 0 )
       {
          operator = XOR;
       }
       else if ( (short) strcmp(token,"+") == 0 )
       {
          operator = ADD;
       }
       else if ( (short )strcmp(token,"-") == 0 )
       {
          operator = SUB;
       }
       else if ( (short) strcmp(token,"&") == 0 )
       {
          operator = AND;
       }
       else if ( (short) strcmp(token,"|") == 0 )
       {
          operator = OR;
       }
       else if ( (short) strcmp(token,"*") == 0 )
       {
          operator = MUL;
       }
       else if ( (short) strcmp(token,"/") == 0 )
       {
          operator = DIV;
       }
       else if ( (short) strcmp(token,"rem") == 0 )
       {
          operator = REM;
       }
       else if ( (short) strcmp(token,";") == 0 )
       {
          operator = ASN;
       }
       else
       {
          printf("unknown operator %s\n",token);
          return(32);
       }

       if ( operator != ASN )  // rslt = left op right ;
       {

          cursor = gettkn( token , cursor );

          if ( token[0] == 0 )
          {
             printf("syntax: <rslt> = <lt> <op> <rt> but <rt> missing\n");
             return(33);
          }

          errcode = parse_dataref();
          if ( errcode != 0 )
          {
             printf("error %d from parse_dataref\n",errcode);
             return(53);
          }


          right_type = result_type;
          right_aregno = aregno;
          right_dregno = dregno;
          right_constant = constant;
          right_offset   = offset;
          right_datatype = datatype;
          right_datalen  = datalen;
          right_indexlen  = indexlen;

          dest_mask = 0x0200;                    // rslt = left op right ;
       }
       else if ( rslt_type == 5 )
       {
          dest_mask = (rslt_dregno<<9);          // dreg = left ;
       }
       else if ( rslt_type == 6 )
       {
          dest_mask = (rslt_aregno<<9) | 0x0040; // areg = left ;
       }
       else
       {
          dest_mask = 0x0200;                    // rslt = left ;
       }


       //
       // now emit code for expression
       // based on rslt_*, left_*, right_*
       // and operator.
       //

         type = left_type;
         areg = left_aregno;
         dreg = left_dregno;
         con  = left_constant;
         dtype = left_datatype;
         dlen = left_datalen;
         ilen = left_indexlen;
         off  = left_offset;
       
         if ( dlen == 1 )
           len_mask = 0x1000;
         else if ( dlen == 2 )
           len_mask = 0x3000;
         else if ( dlen == 4 )
           len_mask = 0x2000;
         else
         {
           printf("internal error: dlen = %d\n",dlen);
           return(34);
         }


         opt_flag = 0;


         if (    operator != ASN 
              && rslt_type == 5
              && left_type == 5
              && rslt_dregno == left_dregno
              && rslt_datalen == right_datalen
              && rslt_datalen == left_datalen    // dreg = dreg op right ;
            )
         {
            opt_flag = 1;
            dest_mask = (rslt_dregno<<9);
            dest_reg  = rslt_dregno;
            goto  do_operation;
         }


//
// left types:
//
//            1   *areg++
//            2   *areg
//            3   areg[constant]
//            4   areg[dreg]
//            5   dreg
//            6   areg
//            7   data
//            8   data[constant]
//            9   data[dreg]
//            10  constant
//            11  &data
//            12  "literal string"
//

         switch( (short) type )
         {
            case 1:   // move.s (an)+,d1
               opword = 0x0018 | dest_mask | len_mask | areg;
               emit2();
               if ( operator == ASN && rslt_type == 6 )
               {
                  if (compile_listing == 1 )
                     printf("movea.%d (a%d)+,a%d\n",
                          dlen,areg,(dest_mask>>9));
               }
               else
               {
                  if (compile_listing == 1 )
                     printf("move.%d  (a%d)+,d%d\n",
                        dlen,areg,(dest_mask>>9));
               }
               break;
       
            case 2:   // move.s (an),d1
               opword = 0x0010 | dest_mask | len_mask | areg;
               emit2();
               if ( operator == ASN && rslt_type == 6 )
               {
                  if (compile_listing == 1 )
                     printf("movea.%d (a%d),a%d\n",
                        dlen,areg,(dest_mask>>9));
               }
               else
               {
                  if (compile_listing == 1 )
                     printf("move.%d  (a%d),d%d\n",
                        dlen,areg,(dest_mask>>9));
               }
               break;
       
            case 3:   // move.s d(an),d1
               opword = 0x0028 | dest_mask | len_mask | areg;
               shortcon = con;
               dispword = shortcon*dlen;
               emit4();
               if (compile_listing == 1 )
               {
                  if ( operator == ASN && rslt_type == 6 )
                      printf("movea.%d %d(a%d),a%d\n",
                             dlen,dispword,areg,(dest_mask>>9));
                  else
                      printf("move.%d  %d(a%d),d%d\n",
                             dlen,dispword,areg,(dest_mask>>9));
               }
               break;
       
            case 4:   // move.s 0(an,di),d1

               opword = 0x0030 | areg | dest_mask | len_mask;
               dispword = 0x0000;  // D/Aiii W/L000 0000 0000
               if ( ilen == 2 )
               {
                 dispword = 0x0000 | ( dreg << 12 );
               }
               else if ( ilen == 4 )
               {
                 dispword = 0x0800 | ( dreg << 12 );
               }
               else
               {
                 printf("internal error: bad ilen = %d\n",ilen);
                 return(35);
               }
               emit4();
               if (compile_listing == 1 )
               {
                  if ( operator == ASN && rslt_type == 6 )
                      printf("movea.%d 0(a%d,d%d.%d),a%d\n",
                             dlen,areg,dreg,ilen,(dest_mask>>9));
	          else
                      printf("move.%d  0(a%d,d%d.%d),d%d\n",
                             dlen,areg,dreg,ilen,(dest_mask>>9));
               }
               break;
       
            case 5:   // move.s dn,d1
               opword = 0x0000 | dest_mask | len_mask | dreg;
               emit2();
               if (compile_listing == 1 )
               {
                  if ( operator == ASN && rslt_type == 6 )
                      printf("movea.%d d%d,a%d\n",
                             dlen,dreg,(dest_mask>>9));
	          else
                      printf("move.%d  d%d,d%d\n",
                             dlen,dreg,(dest_mask>>9));
               }
               break;
       
            case 6:   // move.l an,d1
               opword = 0x2008 | dest_mask | areg ;
               emit2();
               if (compile_listing == 1 )
               {
                  if ( operator == ASN && rslt_type == 6 )
                      printf("movea.4 a%d,a%d\n",
                             areg,(dest_mask>>9));
                  else
                      printf("move.4  a%d,d%d\n",
                             areg,(dest_mask>>9));
               }
               break;
       
            case 7:   // move.s d(a5),d1
               opword = 0x002d | dest_mask | len_mask;
               dispword = off;
               emit4();
               if (compile_listing == 1 )
               {
                  if ( operator == ASN && rslt_type == 6 )
                      printf("movea.%d %d(a5),a%d\n",
                             dlen,dispword,(dest_mask>>9));
	          else
                      printf("move.%d  %d(a5),d%d\n",
                             dlen,dispword,(dest_mask>>9));
               }
               break;
       
            case 8:   // move.s d(a5),d1
               opword = 0x002d | dest_mask | len_mask;
               shortcon = con;
               dispword = off + (shortcon*dlen);
               emit4();
               if (compile_listing == 1 )
               {
                  if ( operator == ASN && rslt_type == 6 )
                      printf("movea.%d %d(a5),a%d\n",
                             dlen,dispword,(dest_mask>>9));
	          else
                      printf("move.%d  %d(a5),d%d\n",
                             dlen,dispword,(dest_mask>>9));
               }
               break;
       
            case 9:   // move.s d(an,di),d1

               opword = 0x0030 | dest_mask | areg | len_mask;

               if ( ilen == 2 )
               {
                 // D/Aiii W/L000 dddd dddd
                 dispword = 0x0000 | ( dreg << 12 ) | off;
               }
               else if ( ilen == 4 )
               {
                 // D/Aiii W/L000 dddd dddd
                 dispword = 0x0800 | ( dreg << 12 ) | off; 
               }
               else
               {
                 printf("internal error: bad ilen = %d\n",ilen);
                 return(36);
               }
               emit4();
               if (compile_listing == 1 )
               {
                  if ( operator == ASN && rslt_type == 6 )
                      printf("movea.%d %d(a%d,d%d.%d),a%d\n",
                             dlen,off,areg,dreg,ilen,(dest_mask>>9));
	          else
                      printf("move.%d  %d(a%d,d%d.%d),d%d\n",
                             dlen,off,areg,dreg,ilen,(dest_mask>>9));
               }
               break;
       
            case 10:   // move.l #imm,d1
               opword = 0x003c | dest_mask | len_mask;
               dispword = con;

               if ( dlen == 2 || dlen == 1 )
               {
                   emit4();
               }
               else if ( dlen == 4 )
               {
                   dispword = con>>16;
                   thirdword = con & 0xffff;
                   emit6();
               }
               if (compile_listing == 1 )
               {
                  if ( operator == ASN && rslt_type == 6 )
                     printf("movea.%d #%ld,a%d\n",
                        dlen,con,(dest_mask>>9));
	          else
                     printf("move.%d  #%ld,d%d\n",
                        dlen,con,(dest_mask>>9));
               }
               break;
       
            case 11:
               if ( operator == ASN && rslt_type == 6 )
               {
                  // lea d(a5),ar
                  opword = 0x41ed | (rslt_aregno<<9);
                  dispword = off;
                  emit4();
                  if (compile_listing == 1 )
                     printf("lea     %d(a5),a%d\n",
                                 off,rslt_aregno);
               }
               else
               {
                  // lea d(a5),a0
                  opword = 0x41ed;
                  dispword = off;
                  emit4();
                  if (compile_listing == 1 )
                     printf("lea     %d(a5),a0\n", off);

                  // move a0,dr
                  opword = 0x2008 | dest_mask;
                  emit2();
                  if (compile_listing == 1 )
                     printf("move a0,d%d\n", (dest_mask>>9) );
                  break;
               }
               break;
       
            default:
               printf("internal error: 3 bad rtype = %d\n",type);
               break;
         }

         if ( operator == ASN && rslt_type == 5 )
              goto done_assign;
            
         if ( operator == ASN && rslt_type == 6 )
              goto done_assign;

         dest_mask = 0x0200;
         dest_reg = 1;

do_operation:

         if (    ( operator == DIV || operator == REM )
              && left_datalen < 4
            )
         {
            // zero upper half of dest_reg
            opword = 0x4840 | dest_reg; // swap  dn
            emit2();
            if (compile_listing == 1 )
               printf("swap   d%d\n",dest_reg);

            opword = 0x4240 | dest_reg; // clr.w dn
            emit2();
            if (compile_listing == 1 )
               printf("clr.w  d%d\n",dest_reg);

            opword = 0x4840 | dest_reg; // swap  dn
            emit2();
            if (compile_listing == 1 )
               printf("swap   d%d\n",dest_reg);
         }


         switch( operator )
         {
             case XOR:
               op_mask = 0xb000;
               strcpy(opname,"eor");
               break;

             case ADD:
               op_mask = 0xd000;
               strcpy(opname,"add");
               break;

             case SUB:
               op_mask = 0x9000;
               strcpy(opname,"sub");
               break;

             case AND:
               op_mask = 0xc000;
               strcpy(opname,"and");
               break;

             case OR:
               op_mask = 0x8000;
               strcpy(opname,"or ");
               break;

             case MUL:
               op_mask = 0xc0c0;
               strcpy(opname,"mul");
               break;

             case DIV:
               op_mask = 0x80c0;
               strcpy(opname,"div");
               break;

             case REM:
               op_mask = 0x80c0;
               strcpy(opname,"div");
               break;

             default:
               op_mask = 0x0000;
               strcpy(opname,"oop");
               break;
         }

	 if ( operator  == ASN )
	     goto do_assign;

         type = right_type;
         areg = right_aregno;
         dreg = right_dregno;
         con  = right_constant;
         dtype = right_datatype;
         dlen = right_datalen;
         ilen = right_indexlen;
         off  = right_offset;
       
         if ( dlen == 1 )
           len_mask = 0x0000;
         else if ( dlen == 2 )
           len_mask = 0x0040;
         else if ( dlen == 4 )
           len_mask = 0x0080;
         else
         {
           printf("internal error: dlen = %d\n",dlen);
           return(37);
         }
//
// right types:
//
//            1   *areg++
//            2   *areg
//            3   areg[constant]
//            4   areg[dreg]
//            5   dreg
//            6   areg
//            7   data
//            8   data[constant]
//            9   data[dreg]
//            10  constant
//            11  &data
//            12  "literal string"
//
       
         switch( (short ) type )
         {
            case 1:   // op.s (an)+,d1
               opword = 0x0018 | op_mask | dest_mask | len_mask | areg;
               emit2();
               if (compile_listing == 1 )
                  printf("%s.%d   (a%d)+,d%d\n",opname,dlen,areg,dest_reg);
               break;
       
            case 2:   // op.s (an),d1
               opword = 0x0010 | op_mask | dest_mask | len_mask | areg;
               emit2();
               if (compile_listing == 1 )
                  printf("%s.%d   (a%d),d%d\n",opname,dlen,areg,dest_reg);
               break;
       
            case 3:   // op.s d(an),d1
               opword = 0x0028 | op_mask | dest_mask | len_mask | areg;
               shortcon = con;
               dispword = shortcon*dlen;
               emit4();
               if (compile_listing == 1 )
                  printf("%s.%d   %d(a%d),d%d\n",
                        opname,dlen,dispword,areg,dest_reg);
               break;
       
            case 4:   // op.s 0(an,di),d1
               opword = 0x0030 | op_mask | dest_mask | areg | len_mask;
               dispword = 0x0000;  // D/Aiii W/L000 0000 0000
               if ( ilen == 2 )
               {
                 dispword = 0x0000 | ( dreg << 12 );
               }
               else if ( ilen == 4 )
               {
                 dispword = 0x0800 | ( dreg << 12 );
               }
               else
               {
                 printf("internal error: bad ilen = %d\n",ilen);
                 return(38);
               }
               emit4();
               if (compile_listing == 1 )
                  printf("%s.%d   0(a%d,d%d.%d),d%d\n",
                        opname,dlen,areg,dreg,ilen,dest_reg);
               break;
       
            case 5:   // op.s dn,d1
               if ( operator == XOR )
               {
                   opword = 0xb100 | (dreg<<9) | len_mask | dest_reg;
               }
               else
               {
                   opword = 0x0000 | op_mask | dest_mask | len_mask | dreg;
               }
               emit2();
               if (compile_listing == 1 )
                  printf("%s.%d   d%d,d%d\n",opname,dlen,dreg,dest_reg);
               break;
       
            case 6:   // op.l an,d1
               opword = 0x2008 | op_mask | areg | dest_mask;
               emit2();
               if (compile_listing == 1 )
                  printf("%s.4   a%d,d%d\n",opname,areg,dest_reg);
               break;
       
            case 7:   // op.s d(a5),d1
               opword = 0x002d | op_mask | dest_mask | len_mask;
               dispword = off;
               emit4();
               if (compile_listing == 1 )
                  printf("%s.%d   %d(a5),d%d\n",
                        opname,dlen,dispword,dest_reg);
               break;
       
            case 8:   // op.s d(a5),d1
               opword = 0x002d | op_mask | len_mask | dest_mask;
               shortcon = con;
               dispword = off + (shortcon*dlen);
               emit4();
               if (compile_listing == 1 )
                  printf("%s.%d   %d(a5),d%d\n",
                      opname,dlen,dispword,dest_reg);
               break;
       
            case 9:   // op.s d(an,di),d1
               opword = 0x0030 | op_mask | dest_mask | areg | len_mask;
               if ( ilen == 2 )
               {
                 // D/Aiii W/L000 dddd dddd
                 dispword = 0x0000 | ( dreg << 12 ) | off;
               }
               else if ( ilen == 4 )
               {
                 // D/Aiii W/L000 dddd dddd
                 dispword = 0x0800 | ( dreg << 12 ) | off; 
               }
               else
               {
                 printf("internal error: bad ilen = %d\n",ilen);
                 return(39);
               }
               emit4();
               if (compile_listing == 1 )
                  printf("%s.%d   %d(a%d,d%d.%d),d%d\n",
                        opname,dlen,off,areg,dreg,ilen,dest_reg);
               break;
       
            case 10:   // opi.s #imm,d1
	       switch( operator )
               {
                 case XOR:
                   op_mask = 0x0a00;
                   opword = op_mask | len_mask | dest_reg;
                   break;
                 case ADD:
                   op_mask = 0x0600;
                   opword = op_mask | len_mask | dest_reg;
                   break;
                 case SUB:
                   op_mask = 0x0400;
                   opword = op_mask | len_mask | dest_reg;
                   break;
                 case AND:
                   op_mask = 0x0200;
                   opword = op_mask | len_mask | dest_reg;
                   break;
                 case OR:
                   op_mask = 0x0000;
                   opword =  0x0000 | len_mask | dest_reg;
                   break;
                 case MUL:
                   dlen = 2;
                   opword = 0xc0fc | dest_mask;
                   break;
                 case DIV:
                   dlen = 2;
                   opword = 0x80fc | dest_mask;
                   break;
                 case REM:
                   dlen = 2;
                   opword = 0x80fc | dest_mask;
                   break;
                 default:
                   op_mask = 0x0000;
                   opword = 0x0001 | op_mask | len_mask;
                   break;
               }

               dispword = con;

                     
               if ( dlen == 2 || dlen == 1 )
               {
                  emit4();
                  if (compile_listing == 1 )
                     printf("%s.%d   #%ld,d%d\n",
                        opname,dlen,con,dest_reg);
               }
               else
               {
                  dispword = con>>16;;
                  thirdword = con & 0xffff;
                  emit6();
                  if (compile_listing == 1 )
                     printf("%s.%d   #%ld,d%d\n",
                         opname,dlen,con,dest_reg);
               }

               break;
       
            case 11:   // lea  d(a5),a0
            case 12:
               opword = 0x41ed;
               dispword = off;
               emit4();
               if (compile_listing == 1 )
	          printf("lea     %d(a5),a0\n",dispword);
               break;
       
            default:
               printf("internal error: 4 bad rtype = %d\n",type);
               break;
         }

         if ( operator == REM )
         {
            opword = 0x4240 | dest_reg; // clr.w dn
            emit2();
            if (compile_listing == 1 )
               printf("clr.w  d%d\n",dest_reg);

            opword = 0x4840 | dest_reg; // swap  dn
            emit2();
            if (compile_listing == 1 )
               printf("swap   d%d\n",dest_reg);
         }
         else if ( operator == DIV )
         {
            opword = 0x4840 | dest_reg; // swap  dn
            emit2();
            if (compile_listing == 1 )
               printf("swap   d%d\n",dest_reg);

            opword = 0x4240 | dest_reg; // clr.w dn
            emit2();
            if (compile_listing == 1 )
               printf("clr.w  d%d\n",dest_reg);

            opword = 0x4840 | dest_reg; // swap  dn
            emit2();
            if (compile_listing == 1 )
               printf("swap   d%d\n",dest_reg);
         }
         else
         {

         }

         if ( opt_flag == 1 )
            goto done_assign;

 do_assign:
	 

         type = rslt_type;
         areg = rslt_aregno;
         dreg = rslt_dregno;
         con  = rslt_constant;
         dtype = rslt_datatype;
         dlen = rslt_datalen;
         ilen = rslt_indexlen;
         off  = rslt_offset;
       
         if ( dlen == 1 )
           len_mask = 0x1000;
         else if ( dlen == 2 )
           len_mask = 0x3000;
         else if ( dlen == 4 )
           len_mask = 0x2000;
         else
         {
           printf("internal error: dlen = %d\n",dlen);
           return(40);
         }
//
// result types:
//
//            1   *areg++
//            2   *areg
//            3   areg[constant]
//            4   areg[dreg]
//            5   dreg
//            6   areg
//            7   data
//            8   data[constant]
//            9   data[dreg]
//            10  constant
//            11  &data
//            12  "literal string"
//
       
         switch( (short ) type )
         {
            case 1:   // move.s d1,(an)+
               opword = 0x00c1 | len_mask | (areg<<9);
               emit2();
               if (compile_listing == 1 )
                  printf("move.%d  d1,(a%d)+\n", dlen,areg);
               break;
       
            case 2:   // move.s d1,(an)
               opword = 0x0081 | len_mask | (areg<<9);
               emit2();
               if (compile_listing == 1 )
                  printf("move.%d  d1,(a%d)\n", dlen,areg);
               break;
       
            case 3:   // move.s d1,d(an)
               opword = 0x0141 | len_mask | (areg<<9);
               shortcon = con;
               dispword = shortcon*dlen;
               emit4();
               if (compile_listing == 1 )
                  printf("move.%d  d1,%d(a%d)\n",
                        dlen,dispword,areg);
               break;
       
            case 4:   // move.s d1,0(an,di)
               opword = 0x0181 | (areg<<9) | len_mask; // then idx word
               dispword = 0x0000;  // D/Aiii W/L000 0000 0000
               if ( ilen == 2 )
               {
                 dispword = 0x0000 | ( dreg << 12 );
               }
               else if ( ilen == 4 )
               {
                 dispword = 0x0800 | ( dreg << 12 );
               }
               else
               {
                 printf("internal error: bad ilen = %d\n",ilen);
                 return(41);
               }
               emit4();
               if (compile_listing == 1 )
                  printf("move.%d  d1,0(a%d,d%d.%d)\n",
                        dlen,areg,dreg,ilen);
               break;
       
            case 5:   // move.s d1,dn
               opword = 0x0001 | len_mask | (dreg<<9);
               emit2();
               if (compile_listing == 1 )
	          printf("move.%d  d1,d%d\n",dlen,dreg);
               break;
       
            case 6:   // movea.l d1,an
               opword = 0x0041 | len_mask | (areg<<9) ;
               emit2();
               if (compile_listing == 1 )
	          printf("movea.4 d1,a%d\n",areg);
               break;
       
            case 7:   // move.s d1,d(a5)
               opword = 0x0b41 | len_mask;
               dispword = off;
               emit4();
               if (compile_listing == 1 )
                   printf("move.%d  d1,%d(a5)\n",dlen,dispword);
               break;
       
            case 8:   // move.s d1,d(a5)
               opword = 0x0b41 | len_mask;
               shortcon = con;
               dispword = off + (shortcon*dlen);
               emit4();
               if ( compile_listing == 1 )
                  printf("move.%d  d1,%d(a5)\n",dlen,dispword);
               break;
       
            case 9:   // move.s d1,d(an,di)
               opword = 0x0181 | (areg<<9) | len_mask;
               if ( ilen == 2 )
               {
                 // D/Aiii W/L000 dddd dddd
                 dispword = 0x0000 | ( dreg << 12 ) | off;
               }
               else if ( ilen == 4 )
               {
                 // D/Aiii W/L000 dddd dddd
                 dispword = 0x0800 | ( dreg << 12 ) | off; 
               }
               else
               {
                 printf("internal error: bad ilen = %d\n",ilen);
                 return(42);
               }
               emit4();
               if ( compile_listing == 1 )
                  printf("move.%d  d1,%d(a%d,d%d.%d)\n",
                        dlen,off,areg,dreg,ilen);
               break;
       
            default:
               printf("internal error: 1 bad rtype = %d\n",type);
               break;
         }

done_assign:

       //
       // end emit code
       //
    }

} /* END OF FOR OVER LINES */

// should never get here

return(9999);

}







char *gettkn( token , inpcursor )
  char *token;
  char *inpcursor;
{
    int i;
    char *charp;
    char sc;
    char sc2;
    char sc3;
    char delimchar;
    
    /********************/
    
    charp = inpcursor;
    i = 0;

    for(;;)  // skip leading whitespace
    {
        sc = charp[0];
        if (sc == ' ' || sc == '\t')
        {
            charp++;
        }
        else
        {
            break;
        }
    }

    if ( sc == ';' )
    {
      token[0] = ';';
      token[1] = 0;
      return( charp+1 );
    }

    if ( sc == '"' )
    {
       token[i++] = sc;
       charp++; 
       delimchar = '"';
    }
    else
    {
       delimchar = ' ';
    }

    if ( sc == 39 )
    {
       token[0] = sc;
       token[1] = charp[1];
       token[2] = charp[2];
       token[3] = 0;
       charp += 3;
       return(charp);
    }


    sc2 = charp[1];
    sc3 = charp[2];

    if (   ( sc == '>' && sc2  == '>' && sc3 == '>' )
        || ( sc == '<' && sc2  == '<' && sc3 == '<' )
       )
    {
      token[0] = sc;
      token[1] = sc;
      token[2] = sc;
      token[3] = 0;
      charp += 3;
      return(charp); 
    }


    if (    ( sc == '+' && sc2  == '+' )
         || ( sc == '-' && sc2  == '-' )
         || ( sc == '=' && sc2  == '=' )
         || ( sc == '!' && sc2  == '=' )
         || ( sc == '&' && sc2  == '&' )
         || ( sc == '|' && sc2  == '|' )
         || ( sc == '>' && sc2  == '=' )
         || ( sc == '>' && sc2  == '>' )
         || ( sc == '<' && sc2  == '=' )
         || ( sc == '<' && sc2  == '<' )
         || ( sc == '+' && sc2  == '=' )
         || ( sc == '-' && sc2  == '=' )
         || ( sc == '&' && sc2  == '=' )
         || ( sc == '|' && sc2  == '=' )
         || ( sc == '%' && sc2  == '=' )
         || ( sc == '/' && sc2  == '=' )
         || ( sc == '*' && sc2  == '=' )
         || ( sc == '^' && sc2  == '=' )
       )
    {
      token[0] = sc;
      token[1] = charp[1];
      token[2] = 0;
      charp++;
      charp++;
      return(charp); 
    }

    if (    sc == '/'
         && ( charp[1] == '/'
              || charp[1] == '*'
            )
       )
    {
      token[0] = 0;
      return(charp); 
    }


    if (  sc == ','|| sc == '+'
       || sc == '-'|| sc == '^'
       || sc == '/'|| sc == '*'
       || sc == '='|| sc == '&'
       || sc == '|'|| sc == '&'
       || sc == '>'|| sc == '<'
       || sc == ')'|| sc == ';'
       || sc == '}'|| sc == '{'
       || sc == '!'
      )
    {
        token[i] = sc;
        token[i+1] = 0;
        charp++;
        return(charp);
    }

    
    for(;;)
    {
        sc = charp[0];

        if ( delimchar == ' ' && sc == '(' )
        {
           token[i] = '(';
           token[i+1] = 0;
           charp++;
           return(charp);
        }

        if ( delimchar == ' ' && sc == '[' )
        {
           delimchar = ']';
        }


        if ( delimchar == ' ' )
        {

          if (   sc == 0
              || sc == 10 || sc == 13
              || sc == ','|| sc == '+'
              || sc == '-'|| sc == '^'
              || sc == '/'|| sc == '*'
              || sc == '='|| sc == '&'
              || sc == '>'|| sc == '<'
              || sc == '|'|| sc == '&'
              || sc == ')'|| sc == ';'
              || sc == '}'|| sc == '{'
             )
          {
              token[i] = 0;
              return(charp);
          }
        }
        
        if (sc == delimchar || sc == '\t')
        {
           if ( delimchar == ' ' )
           {
              charp++;
              token[i] = 0;
              return(charp);
           }
           else
           {
              token[i] = sc;
              token[i+1] = 0;
              charp++;
              return(charp);
           }
        }
        token[i++] = sc;
        charp++; 
    }
}






short int getaregno( char * regname, short *pointeelenp )
{
int i;
int rslt;

   for (i=1;i<nextaddrreg;i++)
   {
      rslt = strcmp( addrregname[i], regname );
      if ( rslt == 0 )
      {
         *pointeelenp = addrregtosize[i];
         return( i );
      }
   }
   return( 0 );
}





short int getdregno( char * regname, short *reglenp )
{
int i;
int rslt;

   for (i=2;i<nextdatareg;i++)
   {
      rslt = strcmp( dataregname[i], regname );
      if ( rslt == 0 )
      {
         *reglenp = dataregsize[i];     // 1,2, or 4
         return( i );
      }
   }
   return( 0 );
}





//
// starts parsing for data/addr ref
// at current token
//
// inputs: token and rest of line
// outputs:
//
//   aregno
//   dregno
//   long constant
//   short datalen
//   long indexlen
//   result_type:
//   ============
//            1   *areg++
//            2   *areg
//            3   areg[constant]
//            4   areg[dreg]
//            5   dreg
//            6   areg
//            7   data
//            8   data[constant]
//            9   data[dreg]
//            10  constant
//            11  &data
//            12  "literal string"
//
//
//
// Note: leaves token as token after *areg++, etc.
//

short int  parse_dataref()
{
char *xp;
char *cp;
int  i;
short pointeelen;
short  reglen;
int newoff;


       aregno = 0;
       result_type = 0;


       if (token[0] == 0)
       {
          printf("no token\n");
          return(60);
       }


       if ( token[0] == '"' )  // "literal"
       {
          // go thru token converting \n to newline

          xp = token+1;
          cp = a5_area+a5_litoff;
          newoff = a5_litoff;
          while ( *xp != 0 && *xp != '"' )
          {
             if ( *xp == '\\' && *(xp+1) == 'n' )  // \n ?
             {
                *cp++ = 10;
                newoff++;
                xp+=2;
             }
             else if ( *xp == '/' )  // % alternate
             {
                *cp++ = '%';
                newoff++;
                xp++;
             }
             else
             {
                *cp++ = *xp++;
                newoff++;
             }
          }
          *cp = 0;
          result_type = 12;

          offset = a5_litoff;
          datalen = 4;

          a5_litoff = newoff+1;
          cursor = gettkn( token , cursor );  // pos at =
	  return(0);
       }
       else if ( (short) strcmp(token,"&") == 0 )  // &data
       {
         cursor = gettkn( token , cursor );

         for (i=0;i<numgbls;i++)
         {
           rslt = strcmp(gblname[i],token);
           if (rslt == 0 )
           {
              offset = gbloffset[i];
              datatype = gbltype[i];
              datalen  = gblsize[i];
              cursor = gettkn( token , cursor );  // pos at =

              result_type = 11;
              datalen = 4;
              return(0);
           }
         }
         printf("syntax: &data is only allowed & form\n");
         return(61);
       }
       else if ( (short) strcmp(token,"*") == 0 )
       {

         cursor = gettkn( token , cursor );
         strcpy( aregname, token );

         aregno = getaregno( aregname, &pointeelen );

         if ( aregno == 0 )
         {
             printf("syntax: *areg...  %s not ptr var = \n",aregname);
             return(62);
         }

         datalen = pointeelen;

         cursor = gettkn( token , cursor );

         if ( (short) strcmp(token,"++") == 0 )
         {
             result_type = 1;
             cursor = gettkn( token , cursor ); // pos at =
         }
         else
         {
             result_type = 2;
         }
       }
       else   // not starting with *
       {

         if (  (cp = strstr(token,"[")) != NULL )
         {
            *cp = 0;
            strcpy( aregname, token );
            aregno = getaregno( token , &pointeelen );
            datalen = pointeelen;

            if ( aregno == 0 )
            {
               for (i=0;i<numgbls;i++)
               {
                 rslt = strcmp(gblname[i],token);
                 if ( rslt == 0 )
                 {
                    offset = gbloffset[i];
                    datatype = gbltype[i];
                    datalen  = gblsize[i];

                    cursor = gettkn( token , cursor );  // pos at =

                    xp = strstr(cp+1,"]");
                    if ( xp != NULL )
                       *xp = 0;
                    dregno = getdregno( cp+1, &reglen );
                    constant = atoi( cp+1 );
                    indexlen = reglen;

                    if ( dregno == 0 )
                    {
                        result_type = 8;
                    }
                    else
                    {
                        result_type = 9;
                    }
                    return(0);
                 }
               }
	       
               printf("syntax: areg[ ... ] =  %s not ptr var \n",token);
               return(63);
            }

            xp = strstr(cp+1,"]");
            if ( xp != NULL )
              *xp = 0;
            dregno = getdregno( cp+1, &reglen );
            constant = atoi( cp+1 );
            indexlen = reglen;

            if ( dregno == 0 )
            {
               result_type = 3;
            }
            else
            {
               result_type = 4;
            }
         }
         else // simple - no []
         {

            dregno = getdregno( token, &reglen );
            datalen = reglen;

            if ( dregno == 0 )
            {
               result_type = 6;
               aregno = getaregno( token , &pointeelen );

               if ( aregno == 0 )
               {
                  for (i=0;i<numgbls;i++)
                  {
                    rslt = strcmp(gblname[i],token);
                    if (rslt == 0 )
                    {
                       result_type = 7;
                       offset = gbloffset[i];
                       datatype = gbltype[i];
                       datalen  = gblsize[i];
                       cursor = gettkn( token , cursor );  // pos at =

	               return(0);
                    }
                  }
                  //
                  // not a global - must be a constant
                  //
                  xp = token;
                  if ( xp[0] == '\'' )
                  {
                    constant = xp[1];
                    datalen = 1;
                  }
                  else
                  {
                     xp = strstr( token, "L" );
                     if ( xp != NULL )
                     {
                        *xp = 0;
                        datalen = 4;
                     }
                     else
                        datalen = 2;
                     constant = strtoul( token, NULL, 0 );
                  }

                  result_type = 10;
                  cursor = gettkn( token , cursor );  // pos at =
		  return(0);
               }
               datalen = 4;
            }
            else
            {
               result_type = 5;
            }
         }

         cursor = gettkn( token , cursor );  // pos at =
       }

//
// end parsing ends at token after *xx++ or whatever
//

   return(0);

} // parse_dataref()





char romcallname[36][10]
= {
"malloc" ,  // 0xA2,  void*  (long)
"free"   ,  // 0xA3,  void   (void*)
"kbhit"  ,  // 0x52,  int    (void)
"ngetchx",  // 0x51,  int    (void)
"gc",       // 0x51,  int    (void)   == ngetchx
"sprintf",  // 0x53,  int    (char*,char*,...)
"GKeyIn" ,  // 0x17E,
"GKeyDown", // 0x17D,
"memcpy" ,  // 0x26A, void*  (void*,void*,long)
"memcmp" ,  // 0x270, long   (void*,void*,long)
"memset" ,  // 0x27C, void*  (void*,int,long)
"strcpy" ,  // 0x26C, char*  (char*,char*)
"strcat" ,  // 0x26E, char*  (char*,char*)
"strcmp" ,  // 0x271, long   (char*,char*)
"strlen" ,  // 0x27E, ulong  (char*)
"strstr" ,  // 0x279, char*  (char*,char*)
"strtok" ,  // 0x27A, char*  (char*,char*)
"moveto" ,  // 0x19d,MoveTo( int x, int y)
"clrscr" ,  // 0x19e,ClrScr( void )
"font",     // 0x18f,FontSetSys( int)
"text",     // 0x1a9,DrawStr( int,int,char *,int attr)
"line",     // 0x1a7,DrawLine( int,int,int,int,int attr)
"pixel",    // 0x1a8,DrawPix( int,int,int attr)
"icon",     // 0x1a6,DrawIcon( int,int,ICON *,int attr)
"lineto",   // 0x19c,LineTo( int x,y )
"attr",     // 0x19a,SetCurAttr( int attr )
"clip",     // 0x19b,SetCurClip( RECT * )
"lines",    // 0x196,DrawMultiLines( int x,y, MULTILINES * )
"rectangle",// 0x189,ScrRectFill( RECT *rect,RECT *clip, int attr )
"triangle", // 0x198,FillTriangle( int x0,y0,x1,y1,x2,y2, RECT*, int attr)
"ellipse",  // 0x192,DrawClipEllipse(int x,y,a,b, RECT *, int attr)
"popnew",   // 0x045,int hndl = PopupNew( char *title, int hgt )
"poptext",  // 0x044,PopupAddText( int hndl, int link, char *txt, uint rval) 
"popdo",    // 0x047,PopupDo( int handle, int x,y, uint start_opt)
"heapfree", // 0x097,HeapFree( int handle );
"" };




unsigned short romcall[36]
= {  0xA2,  0xA3,  0x52,  0x51, 0x51,  0x53,  0x17e,  0x17d, 0x26A,
    0x270, 0x27C, 0x26C, 0x26E, 0x271, 0x27E, 0x279, 0x27A, 0x19d,
    0x19e, 0x18f, 0x1a9, 0x1a7, 0x1a8, 0x1a6, 0x19c, 0x19a, 0x19b,
    0x196, 0x189, 0x198, 0x192, 0x045, 0x044, 0x047, 0x097, 0 };


short int getromcall( char * fname )
{
  int i;
  int rslt;

  for (i=0;i<35;i++)
  {
    rslt = strcmp(romcallname[i],fname);
    if ( rslt == 0 )
    {
        return( romcall[i] );
    }
  }
  printf("invalid rom call %s\n",fname);
  return(0);
}



void emit2()
{
unsigned char *codecharp;
unsigned short *codep;

   if (compile_listing == 1 )
   {
#ifdef CMDLINE_VERSION
      printf("            %04x         ",opword);
#endif
#ifdef TI89_VERSION
      printf("            %04x     ",opword);
#endif
      ngetchx();
   }

   codecharp = codeptr + codeoff;
   codep = ( unsigned short * ) codecharp;
   codep[0] = opword;

   codeoff += 2;
}



void emit4()
{
unsigned char *codecharp;
unsigned short *codep;

   if (compile_listing == 1 )
   {
#ifdef CMDLINE_VERSION
      printf("            %04x%04x     ",opword,dispword);
#endif
#ifdef TI89_VERSION
      printf("            %04x%04x ",opword,dispword);
#endif
      ngetchx();
   }

   codecharp = codeptr + codeoff;
   codep = ( unsigned short * ) codecharp;
   codep[0] = opword;
   codep[1] = dispword;

   codeoff += 4;
}



void emit6()
{
unsigned char *codecharp;
unsigned short *codep;

   if (compile_listing == 1 )
   {
      printf("            %04x%04x%04x ", opword,dispword,thirdword);
      ngetchx();
   }

   codecharp = codeptr + codeoff;
   codep = ( unsigned short * ) codecharp;
   codep[0] = opword;
   codep[1] = dispword;
   codep[2] = thirdword;

   codeoff += 6;
}


